from __future__ import annotations

import datetime

from asn1crypto import tsp


def accuracy_to_python(accuracy: tsp.Accuracy) -> datetime.timedelta:
    delta = datetime.timedelta()
    if not accuracy:
        return delta

    if accuracy["seconds"].native:
        delta += datetime.timedelta(seconds=accuracy["seconds"].native)
    if accuracy["millis"].native:
        delta += datetime.timedelta(milliseconds=accuracy["millis"].native)
    if accuracy["micros"].native:
        delta += datetime.timedelta(microseconds=accuracy["micros"].native)
    return delta



from __future__ import annotations

import hashlib
from typing import Iterable, cast

from asn1crypto import algos

from signify._typing import HashFunction
from signify.exceptions import ParseError

# this list must be in the order of worst to best
ACCEPTED_DIGEST_ALGORITHMS = ("md5", "sha1", "sha256", "sha384", "sha512")


def _get_digest_algorithm(
    algorithm: algos.DigestAlgorithm,
    location: str,
    acceptable: Iterable[str] = ACCEPTED_DIGEST_ALGORITHMS,
) -> HashFunction:
    alg = algorithm["algorithm"].native
    if alg not in acceptable:
        raise ParseError(f"{location} must be one of {list(acceptable)}, not {alg}")
    if algorithm["parameters"].native:
        raise ParseError(f"{location} has parameters set, which is unexpected")
    return cast(HashFunction, getattr(hashlib, alg))



from __future__ import annotations

import datetime
import struct
from typing import Any, cast

from asn1crypto.algos import DigestAlgorithm
from asn1crypto.cms import ContentInfo, ContentType, EncapsulatedContentInfo
from asn1crypto.core import (
    Asn1Value,
    BMPString,
    Integer,
    ObjectIdentifier,
    OctetString,
    Sequence,
    SequenceOf,
    SetOf,
)
from asn1crypto.util import utc_with_dst
from asn1crypto.x509 import Extensions, ExtKeyUsageSyntax, Time

# Based on https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf


class CTLVersion(Integer):  # type: ignore[misc]
    """Version of the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_::

        CTLVersion ::= INTEGER {v1(0)}
    """

    _map = {
        0: "v1",
    }


class SubjectUsage(ExtKeyUsageSyntax):  # type: ignore[misc]
    """Subject usage of the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_::

        SubjectUsage ::= EnhancedKeyUsage
    """


class ListIdentifier(OctetString):  # type: ignore[misc]
    """List identifier of the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_::

         ListIdentifier ::= OCTETSTRING
    """


class SubjectIdentifier(OctetString):  # type: ignore[misc]
    """Subject identifier of the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_::

         SubjectIdentifier ::= OCTETSTRING
    """


class SubjectAttributeType(ObjectIdentifier):  # type: ignore[misc]
    _map = {
        "1.3.6.1.4.1.311.10.11.9": "microsoft_ctl_enhkey_usage",
        "1.3.6.1.4.1.311.10.11.11": "microsoft_ctl_friendly_name",
        "1.3.6.1.4.1.311.10.11.20": "microsoft_ctl_key_identifier",
        "1.3.6.1.4.1.311.10.11.29": "microsoft_ctl_subject_name_md5_hash",
        "1.3.6.1.4.1.311.10.11.83": "microsoft_ctl_root_program_cert_policies",
        "1.3.6.1.4.1.311.10.11.98": "microsoft_ctl_auth_root_sha256_hash",
        "1.3.6.1.4.1.311.10.11.104": "microsoft_ctl_disallowed_filetime",
        "1.3.6.1.4.1.311.10.11.105": "microsoft_ctl_root_program_chain_policies",
        "1.3.6.1.4.1.311.10.11.122": "microsoft_ctl_disallowed_enhkey_usage",
        "1.3.6.1.4.1.311.10.11.126": "microsoft_ctl_not_before_filetime",
        "1.3.6.1.4.1.311.10.11.127": "microsoft_ctl_not_before_enhkey_usage",
    }


class SetOfSpecificOctetString(SetOf):  # type: ignore[misc]
    """Specific implementation of a SetOf OctetString that allows parsing directly as
    a value, or as a sequence, depending on the child type.
    """

    _child_spec = OctetString
    children: Any

    def parse(
        self, spec: type[Asn1Value] | None = None, spec_params: Any = None
    ) -> Any:
        if not spec:
            return self.children
        if issubclass(spec, SequenceOf):
            self.children = [spec.load(child.contents) for child in self]
        else:
            self.children = [spec(contents=child.contents) for child in self]
        return self.children


class CTLString(BMPString):  # type: ignore[misc]
    _encoding = "utf-16-le"

    def __unicode__(self) -> str:
        return cast(str, super().__unicode__().rstrip("\0"))


class FileTime(OctetString):  # type: ignore[misc]
    _epoch = datetime.datetime(1601, 1, 1, tzinfo=datetime.timezone.utc)
    _native: datetime.datetime | None

    @property
    def native(self) -> datetime.datetime | None:
        if self.contents is None or not self.contents:
            return None

        if self._native is None:
            value = struct.unpack("<Q", self.contents)[0]
            self._native = self._epoch + datetime.timedelta(microseconds=value / 10)

        return self._native

    def set(self, value: Any) -> None:
        if isinstance(value, datetime.datetime):
            if not value.tzinfo:
                raise ValueError("Must be timezone aware")
            value = value.astimezone(utc_with_dst)
            value = struct.pack("<Q", int((value - self._epoch).total_seconds() * 100))
        OctetString.set(self, value)
        self._native = None


class SubjectAttribute(Sequence):  # type: ignore[misc]
    """Subject attributes of the trusted subject in the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_.
    """

    _fields = [
        ("type", SubjectAttributeType),
        ("values", SetOfSpecificOctetString),
    ]
    _oid_specs: dict[str, type[Asn1Value]] = {
        "microsoft_ctl_enhkey_usage": ExtKeyUsageSyntax,
        "microsoft_ctl_friendly_name": CTLString,
        # "microsoft_ctl_key_identifier": OctetString,
        # "microsoft_ctl_subject_name_md5_hash": OctetString,
        # "microsoft_ctl_root_program_cert_policies": ExtKeyUsageSyntax,
        # "microsoft_ctl_auth_root_sha256_hash": OctetString,
        "microsoft_ctl_disallowed_filetime": FileTime,
        # "microsoft_ctl_root_program_chain_policies": ExtKeyUsageSyntax,
        "microsoft_ctl_disallowed_enhkey_usage": ExtKeyUsageSyntax,
        "microsoft_ctl_not_before_filetime": FileTime,
        "microsoft_ctl_not_before_enhkey_usage": ExtKeyUsageSyntax,
    }

    def _values_spec(self) -> type[Asn1Value] | None:
        return self._oid_specs.get(self["type"].native, None)

    _spec_callbacks = {"values": _values_spec}


class SubjectAttributes(SetOf):  # type: ignore[misc]
    """Subject attributes of the trusted subject in the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_.
    """

    _child_spec = SubjectAttribute


class TrustedSubject(Sequence):  # type: ignore[misc]
    """Trusted subject in the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_::

         TrustedSubject ::= SEQUENCE{
            subjectIdentifier SubjectIdentifier,
            subjectAttributes Attributes OPTIONAL
         }
    """

    _fields = [
        ("subject_identifier", SubjectIdentifier),
        ("subject_attributes", SubjectAttributes, {"optional": True}),
    ]


class TrustedSubjects(SequenceOf):  # type: ignore[misc]
    """Trusted subjects in the CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_::

        TrustedSubjects ::= SEQUENCE OF TrustedSubject
    """

    _child_spec = TrustedSubject


class CertificateTrustList(Sequence):  # type: ignore[misc]
    """CTL structure.

    Based on `MS-CAESO
    <https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/WinArchive/%5bMS-CAESO%5d.pdf>`_::

         CertificateTrustList ::= SEQUENCE {
             version CTLVersion DEFAULT v1,
             subjectUsage SubjectUsage,
             listIdentifier ListIdentifier OPTIONAL,
             sequenceNumber HUGEINTEGER OPTIONAL,
             ctlThisUpdate ChoiceOfTime,
             ctlNextUpdate ChoiceOfTime OPTIONAL,
             subjectAlgorithm AlgorithmIdentifier,
             trustedSubjects TrustedSubjects OPTIONAL,
             ctlExtensions [0] EXPLICIT Extensions OPTIONAL
         }
    """

    _fields = [
        ("version", CTLVersion, {"default": "v1"}),
        ("subject_usage", SubjectUsage),
        ("list_identifier", ListIdentifier, {"optional": True}),
        ("sequence_number", Integer, {"optional": True}),
        ("ctl_this_update", Time),
        ("ctl_next_update", Time, {"optional": True}),
        ("subject_algorithm", DigestAlgorithm),
        ("trusted_subjects", TrustedSubjects, {"optional": True}),
        ("ctl_extensions", Extensions, {"optional": True, "explicit": 0}),
    ]


# Add CTL to acceptable options
ContentType._map["1.3.6.1.4.1.311.10.1"] = "microsoft_ctl"
ContentInfo._oid_specs["microsoft_ctl"] = EncapsulatedContentInfo._oid_specs[
    "microsoft_ctl"
] = CertificateTrustList



#!/usr/bin/env python

# This is a derivative, modified, work from the verify-sigs project.
# Please refer to the LICENSE file in the distribution for more
# information. Original filename: fingerprinter.py
#
# Parts of this file are licensed as follows:
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module effectively implements the relevant parts of the PECOFF_ documentation
to find the relevant parts of the PE structure.

It is also capable of listing all the certificates in the Certificate Table and find
the certificate with type 0x2. The actual parsing of this certificate is perfomed by
:mod:`signify.authenticode`.

.. _PECOFF: http://www.microsoft.com/whdc/system/platform/firmware/PECOFF.mspx
"""

from __future__ import annotations

import collections
import hashlib
import logging
import os
import struct
from typing import Any, BinaryIO, Iterable, Iterator

from typing_extensions import Literal, TypedDict

from signify import fingerprinter
from signify.asn1.hashing import ACCEPTED_DIGEST_ALGORITHMS
from signify.authenticode import structures
from signify.exceptions import AuthenticodeNotSignedError, SignedPEParseError
from signify.x509 import Certificate

logger = logging.getLogger(__name__)

RelRange = collections.namedtuple("RelRange", "start length")


class ParsedCertTable(TypedDict):
    revision: int
    type: int
    certificate: bytes


class SignedPEFile:
    def __init__(self, file_obj: BinaryIO):
        """A PE file that is to be parsed to find the relevant sections for
        Authenticode parsing.

        :param file_obj: A PE file opened in binary file
        """

        self.file = file_obj

        self.file.seek(0, os.SEEK_END)
        self._filelength = self.file.tell()

    def get_authenticode_omit_sections(self) -> dict[str, RelRange] | None:
        """Returns all ranges of the raw file that are relevant for exclusion for the
        calculation of the hash function used in Authenticode.

        The relevant sections are (as per
        `<http://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/Authenticode_PE.docx>`_,
        chapter Calculating the PE Image Hash):

        * The location of the checksum
        * The location of the entry of the Certificate Table in the Data Directory
        * The location of the Certificate Table.

        :returns: dict if successful, or None if not successful
        """

        try:
            locations = self._parse_pe_header_locations()
        except (SignedPEParseError, struct.error):
            return None
        return {
            k: v
            for k, v in locations.items()
            if k in ["checksum", "datadir_certtable", "certtable"]
        }

    def _parse_pe_header_locations(self) -> dict[str, RelRange]:
        """Parses a PE file to find the sections to exclude from the AuthentiCode hash.

        See http://www.microsoft.com/whdc/system/platform/firmware/PECOFF.mspx for
        information about the structure.
        """

        location = {}

        # Check if file starts with MZ
        self.file.seek(0, os.SEEK_SET)
        if self.file.read(2) != b"MZ":
            raise SignedPEParseError("MZ header not found")

        # Offset to e_lfanew (which is the PE header) is at 0x3C of the MZ header
        self.file.seek(0x3C, os.SEEK_SET)
        pe_offset = struct.unpack("<I", self.file.read(4))[0]
        if pe_offset >= self._filelength:
            raise SignedPEParseError(
                "PE header location is beyond file boundaries"
                f"({pe_offset} >= {self._filelength})"
            )

        # Check if the PE header is PE
        self.file.seek(pe_offset, os.SEEK_SET)
        if self.file.read(4) != b"PE\0\0":
            raise SignedPEParseError("PE header not found")

        # The COFF header contains the size of the optional header
        self.file.seek(pe_offset + 20, os.SEEK_SET)
        optional_header_size = struct.unpack("<H", self.file.read(2))[0]
        optional_header_offset = pe_offset + 24
        if optional_header_size + optional_header_offset > self._filelength:
            # This is not strictly a failure for windows, but such files better
            # be treated as generic files. They can not be carrying SignedData.
            raise SignedPEParseError(
                f"The optional header exceeds the file length ({optional_header_size} "
                f"+ {optional_header_offset} > {self._filelength})"
            )

        if optional_header_size < 68:
            # We can't do authenticode-style hashing. If this is a valid binary,
            # which it can be, the header still does not even contain a checksum.
            raise SignedPEParseError(
                f"The optional header size is {optional_header_size} < 68, "
                f"which is insufficient for authenticode",
            )

        # The optional header contains the signature of the image
        self.file.seek(optional_header_offset, os.SEEK_SET)
        signature = struct.unpack("<H", self.file.read(2))[0]
        if signature == 0x10B:  # IMAGE_NT_OPTIONAL_HDR32_MAGIC
            rva_base = optional_header_offset + 92  # NumberOfRvaAndSizes
            cert_base = optional_header_offset + 128  # Certificate Table
        elif signature == 0x20B:  # IMAGE_NT_OPTIONAL_HDR64_MAGIC
            rva_base = optional_header_offset + 108  # NumberOfRvaAndSizes
            cert_base = optional_header_offset + 144  # Certificate Table
        else:
            # A ROM image or such, not in the PE/COFF specs. Not sure what to do.
            raise SignedPEParseError(
                "The PE Optional Header signature is %x, which is unknown", signature
            )

        # According to the specification, the checksum should not be hashed.
        location["checksum"] = RelRange(optional_header_offset + 64, 4)

        # Read the RVA
        if optional_header_offset + optional_header_size < rva_base + 4:
            logger.debug(
                "The PE Optional Header size can not accommodate for the"
                " NumberOfRvaAndSizes field"
            )
            return location
        self.file.seek(rva_base, os.SEEK_SET)
        number_of_rva = struct.unpack("<I", self.file.read(4))[0]
        if number_of_rva < 5:
            logger.debug(
                "The PE Optional Header does not have a Certificate Table entry in its"
                " Data Directory; NumberOfRvaAndSizes = %d",
                number_of_rva,
            )
            return location
        if optional_header_offset + optional_header_size < cert_base + 8:
            logger.debug(
                "The PE Optional Header size can not accommodate for a Certificate"
                " Table entry in its Data Directory"
            )
            return location

        # According to the spec, the certificate table entry of the data directory
        # should be omitted
        location["datadir_certtable"] = RelRange(cert_base, 8)

        # Read the certificate table entry of the Data Directory
        self.file.seek(cert_base, os.SEEK_SET)
        address, size = struct.unpack("<II", self.file.read(8))

        if not size:
            logger.debug("The Certificate Table is empty")
            return location

        if (
            address < optional_header_size + optional_header_offset
            or address + size > self._filelength
        ):
            logger.debug(
                "The location of the Certificate Table in the binary makes no sense and"
                " is either beyond the boundaries of the file, or in the middle of the"
                " PE header; VirtualAddress: %x, Size: %x",
                address,
                size,
            )
            return location

        location["certtable"] = RelRange(address, size)
        return location

    def _parse_cert_table(self) -> Iterator[ParsedCertTable]:
        """Parses the Certificate Table, iterates over all certificates"""

        locations = self.get_authenticode_omit_sections()
        if not locations or "certtable" not in locations:
            raise SignedPEParseError(
                "The PE file does not contain a certificate table."
            )

        position = locations["certtable"].start
        certtable_end = sum(locations["certtable"])
        while position < certtable_end:
            # check if this position is viable, we need at least 8 bytes for our header
            if position + 8 > self._filelength:
                raise SignedPEParseError(
                    "Position of certificate table is beyond length of file"
                )
            self.file.seek(position, os.SEEK_SET)
            length = struct.unpack("<I", self.file.read(4))[0]
            revision = struct.unpack("<H", self.file.read(2))[0]
            certificate_type = struct.unpack("<H", self.file.read(2))[0]

            # check if we are not going to perform a negative read (and 0 bytes is
            # weird as well)
            if length <= 8 or position + length > certtable_end:
                raise SignedPEParseError("Invalid length in certificate table header")
            certificate = self.file.read(length - 8)

            yield {
                "revision": revision,
                "type": certificate_type,
                "certificate": certificate,
            }
            position += length + (8 - length % 8) % 8

    def get_fingerprinter(self) -> fingerprinter.AuthenticodeFingerprinter:
        """Returns a fingerprinter object for this file.

        :rtype: signify.fingerprinter.AuthenticodeFingerprinter
        """
        return fingerprinter.AuthenticodeFingerprinter(self.file)

    @property
    def signed_datas(self) -> Iterator[structures.AuthenticodeSignedData]:
        """Returns an iterator over :class:`AuthenticodeSignedData` objects relevant for
        this PE file. See :meth:`iter_signed_datas`
        """

        yield from self.iter_signed_datas()

    def iter_signed_datas(
        self, *, include_nested: bool = True, ignore_parse_errors: bool = True
    ) -> Iterator[structures.AuthenticodeSignedData]:
        """Returns an iterator over :class:`AuthenticodeSignedData` objects relevant
        for this PE file.

        :param include_nested: Boolean, if True, will also iterate over all nested
            SignedData structures
        :param ignore_parse_errors: Indicates how to handle
            :exc:`SignedPEParseError` that may be raised while fetching
            embedded :class:`structures.AuthenticodeSignedData` structures.

            When :const:`True`,  which is the default and seems to be how Windows
            handles this as well, this will fetch as many valid
            :class:`structures.AuthenticodeSignedData` structures until an exception
            occurs.

            Note that this will also silence the :exc:`SignedPEParseError` that occurs
            when there's no valid :class:`AuthenticodeSignedData` to fetch.

            When :const:`False`, this will raise the :exc:`SignedPEParseError` as
            soon as one occurs.
        :raises SignedPEParseError: For parse errors in the PEFile
        :raises signify.authenticode.AuthenticodeParseError: For parse errors in the
            SignedData
        :return: iterator of signify.authenticode.SignedData
        """

        def recursive_nested(
            signed_data: structures.AuthenticodeSignedData,
        ) -> Iterator[structures.AuthenticodeSignedData]:
            yield signed_data
            if include_nested:
                for nested in signed_data.signer_info.nested_signed_datas:
                    yield from recursive_nested(nested)

        try:
            found = False
            for certificate in self._parse_cert_table():
                if certificate["revision"] != 0x200:
                    raise SignedPEParseError(
                        f"Unknown certificate revision {certificate['revision']!r}"
                    )

                if certificate["type"] == 2:
                    yield from recursive_nested(
                        structures.AuthenticodeSignedData.from_envelope(
                            certificate["certificate"], pefile=self
                        )
                    )
                    found = True

            if not found:
                raise SignedPEParseError(
                    "A SignedData structure was not found in the PE file's Certificate"
                    " Table"
                )
        except SignedPEParseError:
            if not ignore_parse_errors:
                raise

    def _calculate_expected_hashes(
        self,
        signed_datas: Iterable[structures.AuthenticodeSignedData],
        expected_hashes: dict[str, bytes] | None = None,
    ) -> dict[str, bytes]:
        """Calculates the expected hashes that are needed for verification. This
        provides a small speed-up by pre-calculating all hashes, so that not each
        individual SignerInfo object is responsible for calculating their own hash.

        :param signed_datas: The signed datas of this object. Provided to allow
            :meth:`verify` to prefetch these
        :param expected_hashes: Hashes provided by the caller of :meth:`verify`
        :return: All required hashes
        """

        if expected_hashes is None:
            expected_hashes = {}

        # Calculate which hashes we require for the signedinfos
        digest_algorithms = set()
        for signed_data in signed_datas:
            digest_algorithms.add(signed_data.digest_algorithm)

        # Calculate which hashes are needed
        provided_hashes = {getattr(hashlib, t) for t in expected_hashes}
        needed_hashes = digest_algorithms - provided_hashes

        # Calculate the needed hashes
        if needed_hashes:
            fingerprinter = self.get_fingerprinter()
            fingerprinter.add_authenticode_hashers(*needed_hashes)
            expected_hashes.update(fingerprinter.hashes()["authentihash"])

        return expected_hashes

    def verify(
        self,
        *,
        multi_verify_mode: Literal["any", "first", "all", "best"] = "any",
        expected_hashes: dict[str, bytes] | None = None,
        ignore_parse_errors: bool = True,
        **kwargs: Any,
    ) -> list[tuple[structures.AuthenticodeSignedData, Iterable[list[Certificate]]]]:
        """Verifies the SignedData structures. This is a little bit more efficient than
        calling all verify-methods separately.

        :param expected_hashes: When provided, should be a mapping of hash names to
            digests. This could speed up the verification process.
        :param multi_verify_mode: Indicates how to verify when there are multiple
            :class:`structures.AuthenticodeSignedData` objects in this PE file. Can be:

            * 'any' (default) to indicate that any of the signatures must validate
              correctly.
            * 'first' to indicate that the first signature must verify correctly
              (the default of tools such as sigcheck.exe)
            * 'all' to indicate that all signatures must verify
            * 'best' to indicate that the signature using the best hashing algorithm
              must verify (e.g. if both SHA-1 and SHA-256 are present, only SHA-256
              is checked); if multiple signatures exist with the same algorithm,
              any may verify

            This argument has no effect when only one signature is present.
        :param ignore_parse_errors: Indicates how to handle :exc:`SignedPEParseError`
            that may be raised during parsing of the PE file's certificate table.

            When :const:`True`, which is the default and seems to be how Windows
            handles this as well, this will verify based on all available
            :class:`structures.AuthenticodeSignedData` before a parse error occurs.

            :exc:`AuthenticodeNotSignedError` will be raised when no valid
            :class:`structures.AuthenticodeSignedData` exists.

            When :const:`False`, this will raise the :exc:`SignedPEParseError` as soon
            as one occurs. This often occurs before :exc:`AuthenticodeNotSignedError`
            is potentially raised.
        :return: the used structure(s) in validation, as a list of tuples, in the form
            (signed data object, certificate chain)
        :raises AuthenticodeVerificationError: when the verification failed
        :raises SignedPEParseError: for parse errors in the PEFile
        """

        # we need to iterate it twice, so we need to prefetch all signed_datas
        signed_datas = list(
            self.iter_signed_datas(ignore_parse_errors=ignore_parse_errors)
        )

        # if there are no signed_datas, the binary is not signed
        if not signed_datas:
            raise AuthenticodeNotSignedError("No valid SignedData structure was found.")

        # only consider the first signed_data; by selecting it here we prevent
        # calculating more hashes than needed
        if multi_verify_mode == "first":
            signed_datas = [signed_datas[0]]
        elif multi_verify_mode == "best":
            # ACCEPTED_DIGEST_ALGORITHMS contains the algorithms in worst to best order
            best_algorithm = max(
                (sd.digest_algorithm for sd in signed_datas),
                key=lambda alg: [
                    getattr(hashlib, alg) for alg in ACCEPTED_DIGEST_ALGORITHMS
                ].index(alg),
            )
            signed_datas = [
                sd for sd in signed_datas if sd.digest_algorithm == best_algorithm
            ]

        expected_hashes = self._calculate_expected_hashes(signed_datas, expected_hashes)

        # Now iterate over all SignedDatas
        chains = []
        last_error = None
        assert signed_datas
        for signed_data in signed_datas:
            try:
                chains.append(
                    (
                        signed_data,
                        signed_data.verify(
                            expected_hash=expected_hashes[
                                signed_data.digest_algorithm().name
                            ],
                            **kwargs,
                        ),
                    )
                )
            except Exception as e:  # noqa: PERF203
                # best and any are interpreted as any; first doesn't matter either way,
                # but raising where it is raised is a little bit clearer
                if multi_verify_mode in ("all", "first"):
                    raise
                last_error = e
            else:
                if multi_verify_mode not in ("all", "first"):
                    # only return the last one, as we are in mode any/best
                    return chains[-1:]
        if last_error is None:
            return chains
        raise last_error

    def explain_verify(
        self, *args: Any, **kwargs: Any
    ) -> tuple[structures.AuthenticodeVerificationResult, Exception | None]:
        """This will return a value indicating the signature status of this PE file.
        This will not raise an error when the verification fails, but rather
        indicate this through the resulting enum

        :rtype: (signify.authenticode.AuthenticodeVerificationResult, Exception)
        :returns: The verification result, and the exception containing
            more details (if available or None)
        """

        return structures.AuthenticodeVerificationResult.call(
            self.verify, *args, **kwargs
        )



# This is a derivative, modified, work from the verify-sigs project.
# Please refer to the LICENSE file in the distribution for more
# information. Original filename: asn1/spc.py
#
# Parts of this file are licensed as follows:
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Authenticode-specific ASN.1 data structures, called
Software Publishing Certificate (SPC).
"""

from __future__ import annotations

from asn1crypto.algos import DigestInfo
from asn1crypto.cms import (
    CMSAttribute,
    CMSAttributeType,
    ContentInfo,
    ContentType,
    EncapsulatedContentInfo,
)
from asn1crypto.core import (
    Any,
    Asn1Value,
    BitString,
    BMPString,
    Boolean,
    Choice,
    IA5String,
    ObjectIdentifier,
    OctetString,
    Sequence,
    SequenceOf,
    SetOf,
)
from asn1crypto.x509 import Extension, ExtensionId

# based on https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx


class SpcUuid(OctetString):  # type: ignore[misc]
    """SpcUuid.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcUuid ::= OCTETSTRING
    """


class SpcSerializedObject(Sequence):  # type: ignore[misc]
    """SpcSerializedObject.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcSerializedObject ::= SEQUENCE {
            classId SpcUuid,
            serializedData OCTETSTRING
        }
    """

    _fields = [
        ("class_id", SpcUuid),
        ("serialized_data", OctetString),
    ]


class SpcString(Choice):  # type: ignore[misc]
    """SpcString.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcString ::= CHOICE {
            unicode [0] IMPLICIT BMPSTRING,
            ascii [1] IMPLICIT IA5STRING
        }
    """

    _alternatives = [
        ("unicode", BMPString, {"implicit": 0}),
        ("ascii", IA5String, {"implicit": 1}),
    ]


class SpcLink(Choice):  # type: ignore[misc]
    """SpcLink.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcLink ::= CHOICE {
            url [0] IMPLICIT IA5STRING,
            moniker [1] IMPLICIT SpcSerializedObject,
            file [2] EXPLICIT SpcString
        }
    """

    _alternatives = [
        ("url", IA5String, {"implicit": 0}),
        ("moniker", SpcSerializedObject, {"implicit": 1}),
        ("file", SpcString, {"explicit": 2}),
    ]


class SpcImage(Sequence):  # type: ignore[misc]
    """SpcImage.

    Based on the SPC_IMAGE struct in WinTrust.h, e.g. at
    https://microsoft.github.io/windows-docs-rs/doc/windows/Win32/Security/WinTrust/struct.SPC_IMAGE.html
    """

    _fields = [
        ("image_link", SpcLink, {"explicit": 0, "optional": True}),
        ("bitmap", OctetString, {"implicit": 1, "optional": True}),
        ("metafile", OctetString, {"implicit": 2, "optional": True}),
        ("enhanced_metafile", OctetString, {"implicit": 3, "optional": True}),
        ("gif_file", OctetString, {"implicit": 4, "optional": True}),
    ]


class SpcPeImageFlags(BitString):  # type: ignore[misc]
    """SpcPeImageFlags.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcPeImageFlags ::= BIT STRING {
            includeResources            (0),
            includeDebugInfo            (1),
            includeImportAddressTable   (2)
        }

    """

    _map = {
        0: "include_resources",
        1: "include_debug_info",
        2: "include_import_address_table",
    }


class SpcPeImageData(Sequence):  # type: ignore[misc]
    """SpcPeImageData.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcPeImageData ::= SEQUENCE {
            flags SpcPeImageFlags DEFAULT { includeResources },
            file SpcLink
        }

    Note that although this is not in the spec, it is actually explicitly tagged.
    And although it is not optional in the spec, it is actually optional as shown in
    the accompanying text. It is possible that the specs for
    ``SpcAttributeTypeAndOptionalValue.value`` and ``SpcPeImageData.file`` were
    accidentally flipped.
    """

    _fields = [
        ("flags", SpcPeImageFlags, {"default": {"include_resources"}}),
        ("file", SpcLink, {"optional": True, "explicit": 0}),
    ]


class SpcAttributeType(ObjectIdentifier):  # type: ignore[misc]
    """Specific attribute type of a SPC attribute."""

    _map: dict[str, str] = {
        "1.3.6.1.4.1.311.2.1.15": "microsoft_spc_pe_image_data",
    }


class SpcAttributeTypeAndOptionalValue(Sequence):  # type: ignore[misc]
    """Attribute type and optional value.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcAttributeTypeAndOptionalValue ::= SEQUENCE {
            type ObjectID,
            value [0] EXPLICIT ANY OPTIONAL
        }

    Note that although the spec defines this value as explicitly tagged, that's not
    actually the case. It is possible that the specs for
    `SpcAttributeTypeAndOptionalValue.value`` and ``SpcPeImageData.file`` were
    accidentally flipped.
    """

    _fields = [
        ("type", SpcAttributeType),
        ("value", Any, {"optional": True}),
    ]

    _oid_pair = ("type", "value")
    _oid_specs: dict[str, type[Asn1Value]] = {
        "microsoft_spc_pe_image_data": SpcPeImageData,
    }


class SpcIndirectDataContent(Sequence):  # type: ignore[misc]
    """Indirect data content.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcIndirectDataContent ::= SEQUENCE {
            data SpcAttributeTypeAndOptionalValue,
            messageDigest DigestInfo
        }

    Note: although DigestInfo is explicitly defined in the docs, it is simply a copy of
    the RFC DigestInfo.
    """

    _fields = [
        ("data", SpcAttributeTypeAndOptionalValue),
        ("message_digest", DigestInfo),
    ]


class SpcSpOpusInfo(Sequence):  # type: ignore[misc]
    """SpcSpOpusInfo.

    Based on `Windows Authenticode Portable Executable Signature Format
    <https://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/authenticode_pe.docx>`_::

        SpcSpOpusInfo ::= SEQUENCE {
            programName [0] EXPLICIT SpcString OPTIONAL,
            moreInfo [1] EXPLICIT SpcLink OPTIONAL,
        }

    In WinTrust.h, the value pPublisherInfo is also defined. See
    https://microsoft.github.io/windows-docs-rs/doc/windows/Win32/Security/WinTrust/struct.SPC_SP_OPUS_INFO.html
    """

    _fields = [
        ("program_name", SpcString, {"optional": True, "explicit": 0}),
        ("more_info", SpcLink, {"optional": True, "explicit": 1}),
        ("publisher_info", SpcLink, {"optional": True, "explicit": 2}),
    ]


class SetOfSpcSpOpusInfo(SetOf):  # type: ignore[misc]
    _child_spec = SpcSpOpusInfo


class SpcStatementTypeIdentifier(ObjectIdentifier):  # type: ignore[misc]
    _map: dict[str, str] = {
        "1.3.6.1.4.1.311.2.1.21": "microsoft_spc_individual_sp_key_purpose",
        "1.3.6.1.4.1.311.2.1.22": "microsoft_spc_commercial_sp_key_purpose",
    }


class SpcStatementType(SequenceOf):  # type: ignore[misc]
    """SpcStatementType.

    Based on `MS-OSHARED
    <https://learn.microsoft.com/en-us/openspecs/office_file_formats/ms-oshared/c541e3e4-3d04-4cbd-82db-1b73163427fd>`_::

        SpcStatementType ::= SEQUENCE of OBJECT IDENTIFIER

    """

    _child_spec = SpcStatementTypeIdentifier


class SetOfSpcStatementType(SetOf):  # type: ignore[misc]
    _child_spec = SpcStatementType


ContentType._map["1.3.6.1.4.1.311.2.1.4"] = "microsoft_spc_indirect_data_content"
EncapsulatedContentInfo._oid_specs["microsoft_spc_indirect_data_content"] = (
    ContentInfo._oid_specs["microsoft_spc_indirect_data_content"]
) = SpcIndirectDataContent

CMSAttributeType._map["1.3.6.1.4.1.311.2.1.11"] = "microsoft_spc_statement_type"
CMSAttributeType._map["1.3.6.1.4.1.311.2.1.12"] = "microsoft_spc_sp_opus_info"
CMSAttribute._oid_specs["microsoft_spc_sp_opus_info"] = SetOfSpcSpOpusInfo
CMSAttribute._oid_specs["microsoft_spc_statement_type"] = SetOfSpcStatementType


# reverse-engineered certificate extensions


class SpcSpAgencyInformation(Sequence):  # type: ignore[misc]
    """Reverse-engineered extension for certificates, indicating certain information
    on certificate policies. Based on
    https://microsoft.github.io/windows-docs-rs/doc/windows/Win32/Security/WinTrust/struct.SPC_SP_AGENCY_INFO.html

    See also
    https://sotharo-meas.medium.com/cve-2019-1388-windows-privilege-escalation-through-uac-22693fa23f5f
    """

    _fields = [
        ("policy_information", SpcLink, {"explicit": 0}),
        ("policy_display_text", SpcString, {"optional": True, "explicit": 1}),
        ("logo_image", SpcImage, {"optional": True, "implicit": 2}),
        ("logo_link", SpcLink, {"optional": True, "explicit": 3}),
    ]


class SpcFinancialCriteria(Sequence):  # type: ignore[misc]
    """Reverse-engineered extension for certificates"""

    _fields = [
        ("financial_info_available", Boolean, {"default": False}),
        ("meets_criteria", Boolean, {"default": False}),
    ]


ExtensionId._map["1.3.6.1.4.1.311.2.1.10"] = "microsoft_spc_sp_agency_info"
Extension._oid_specs["microsoft_spc_sp_agency_info"] = SpcSpAgencyInformation
ExtensionId._map["1.3.6.1.4.1.311.2.1.27"] = "microsoft_spc_financial_criteria"
Extension._oid_specs["microsoft_spc_financial_criteria"] = SpcFinancialCriteria



from __future__ import annotations

import datetime
from typing import Any, Iterable, cast

from asn1crypto import cms
from asn1crypto.core import Asn1Value
from typing_extensions import Literal

from signify._typing import HashFunction
from signify.asn1.hashing import _get_digest_algorithm
from signify.exceptions import (
    SignerInfoParseError,
    SignerInfoVerificationError,
    VerificationError,
)
from signify.pkcs7 import signeddata
from signify.x509 import VerificationContext
from signify.x509.certificates import Certificate, CertificateName


class SignerInfo:
    """The SignerInfo class is defined in RFC2315 and RFC5652 (amongst others) and
    defines the per-signer information in a :class:`SignedData` structure.

    It is based on the following ASN.1 object (as per RFC2315)::

        SignerInfo ::= SEQUENCE {
          version Version,
          issuerAndSerialNumber IssuerAndSerialNumber,
          digestAlgorithm DigestAlgorithmIdentifier,
          authenticatedAttributes [0] IMPLICIT Attributes OPTIONAL,
          digestEncryptionAlgorithm DigestEncryptionAlgorithmIdentifier,
          encryptedDigest EncryptedDigest,
          unauthenticatedAttributes [1] IMPLICIT Attributes OPTIONAL
        }

    This class supports RFC2315 and RFC5652.

    .. attribute:: data

       The underlying ASN.1 data object

    .. attribute:: parent

       The parent :class:`SignedData` object (or if other SignerInfos are present, it
       may be another object)

    .. attribute:: issuer
       :type: CertificateName

       The issuer of the SignerInfo, i.e. the certificate of the signer of the
       SignedData object.

    .. attribute:: serial_number

       The serial number as specified by the issuer.

    .. attribute:: digest_algorithm

       The digest algorithm, i.e. the hash algorithm, under which the content and the
       authenticated attributes are
       signed.

    .. attribute:: authenticated_attributes
                   unauthenticated_attributes

       A SignerInfo object can contain both signed and unsigned attributes. These
       contain additional information about the signature, but also the content type
       and message digest. The difference between signed and unsigned is that unsigned
       attributes are not validated.

       The type of this attribute is a dictionary. You should not need to access this
       value directly, rather using one of the attributes listed below.

    .. attribute:: digest_encryption_algorithm

       This is the algorithm used for signing the digest with the signer's key.

    .. attribute:: encrypted_digest

       The result of encrypting the message digest and associated information with the
       signer's private key.


    The following attributes are automatically parsed and added to the list of
    attributes if present.

    .. attribute:: message_digest

       This is an authenticated attribute, containing the signed digest of the data.

    .. attribute:: content_type

       This is an authenticated attribute, containing the content type of the content
       being signed.

    .. attribute:: signing_time

       This is an authenticated attribute, containing the timestamp of signing. Note
       that this should only be present in countersigner objects.

    .. attribute:: countersigner

       This is an unauthenticated attribute, containing the countersigner of the
       SignerInfo.

    """

    issuer: CertificateName
    serial_number: int
    authenticated_attributes: dict[str, list[Any]]
    unauthenticated_attributes: dict[str, list[Any]]
    digest_encryption_algorithm: str
    encrypted_digest: bytes
    digest_algorithm: HashFunction
    message_digest: bytes | None
    content_type: str | None
    signing_time: datetime.datetime | None
    countersigner: CounterSignerInfo | None

    _countersigner_class: type[CounterSignerInfo] | str | None = "CounterSignerInfo"
    _required_authenticated_attributes: Iterable[str] = (
        "content_type",
        "message_digest",
    )
    _expected_content_type: str | None = None

    def __init__(
        self, data: cms.SignerInfo, parent: signeddata.SignedData | None = None
    ):
        """
        :param data: The ASN.1 structure of the SignerInfo.
        :param parent: The parent :class:`SignedData` object.
        """
        if isinstance(self._countersigner_class, str):
            self._countersigner_class = globals()[self._countersigner_class]

        self.data = data
        self.parent = parent
        self._parse()

    def _parse(self) -> None:
        if self.data["sid"].name == "subject_key_identifier":
            raise SignerInfoParseError(
                "Cannot handle SignerInfo.sid with a subject_key_identifier"
            )

        self.issuer = CertificateName(self.data["sid"].chosen["issuer"])
        self.serial_number = self.data["sid"].chosen["serial_number"].native
        self.authenticated_attributes = self._parse_attributes(
            self.data["signed_attrs"],
            required=self._required_authenticated_attributes,
        )
        self._encoded_authenticated_attributes = self._encode_attributes(
            self.data["signed_attrs"]
        )
        self.unauthenticated_attributes = self._parse_attributes(
            self.data["unsigned_attrs"]
        )
        self.digest_encryption_algorithm = self.data["signature_algorithm"][
            "algorithm"
        ].native
        self.encrypted_digest = self.data["signature"].native
        self.digest_algorithm = _get_digest_algorithm(
            self.data["digest_algorithm"], location="SignerInfo.digestAlgorithm"
        )

        # Parse the content of the authenticated attributes
        # - The messageDigest
        self.message_digest = None
        if "message_digest" in self.authenticated_attributes:
            if len(self.authenticated_attributes["message_digest"]) != 1:
                raise SignerInfoParseError(
                    "Only one Digest expected in SignerInfo.authenticatedAttributes"
                )

            self.message_digest = self.authenticated_attributes["message_digest"][
                0
            ].native

        # - The contentType
        self.content_type = None
        if "content_type" in self.authenticated_attributes:
            if len(self.authenticated_attributes["content_type"]) != 1:
                raise SignerInfoParseError(
                    "Only one ContentType expected in"
                    " SignerInfo.authenticatedAttributes"
                )

            self.content_type = self.authenticated_attributes["content_type"][0].native

            if (
                self._expected_content_type is not None
                and self.content_type != self._expected_content_type
            ):
                raise SignerInfoParseError(
                    "Unexpected content type for SignerInfo, expected"
                    f" {self._expected_content_type}, got"
                    f" {self.content_type}"
                )

        # - The signingTime (used by countersigner)
        self.signing_time = None
        if "signing_time" in self.authenticated_attributes:
            if len(self.authenticated_attributes["signing_time"]) != 1:
                raise SignerInfoParseError(
                    "Only one SigningTime expected in"
                    " SignerInfo.authenticatedAttributes"
                )

            self.signing_time = self.authenticated_attributes["signing_time"][0].native

        # - The countersigner
        self.countersigner = None
        if "counter_signature" in self.unauthenticated_attributes:
            if len(self.unauthenticated_attributes["counter_signature"]) != 1:
                raise SignerInfoParseError(
                    "Only one CountersignInfo expected in"
                    " SignerInfo.unauthenticatedAttributes"
                )

            assert self._countersigner_class is not None and not isinstance(
                self._countersigner_class, str
            )  # typing
            self.countersigner = self._countersigner_class(
                self.unauthenticated_attributes["counter_signature"][0]
            )

    def check_message_digest(self, data: bytes) -> bool:
        """Given the data, returns whether the hash_algorithm and message_digest match
        the data provided.
        """

        auth_attr_hash = self.digest_algorithm()
        auth_attr_hash.update(data)
        return auth_attr_hash.digest() == self.message_digest

    @classmethod
    def _parse_attributes(
        cls, data: cms.CMSAttributes, required: Iterable[str] = ()
    ) -> dict[str, list[Asn1Value]]:
        """Given a set of Attributes, parses them and returns them as a dict

        :param data: The authenticatedAttributes or unauthenticatedAttributes to process
        :param required: A list of required attributes
        """

        result = {attr["type"].native: list(attr["values"]) for attr in data}

        if not all(x in result for x in required):
            raise SignerInfoParseError(
                "Not all required attributes found."
                f" Required: {required};"
                f" Found: {result}"
            )

        return result

    @classmethod
    def _encode_attributes(cls, data: cms.CMSAttributes) -> bytes:
        """Given a set of Attributes, prepares them for creating a digest. It as per
        RFC 5652 section 5.2, this changes the tag from implicit to explicit.

        :param data: The attributes to encode
        """

        new_attrs = type(data)(contents=data.contents)
        return cast(bytes, new_attrs.dump())

    def _verify_issuer_signature(
        self, issuer: Certificate, context: VerificationContext
    ) -> None:
        """Check the issuer signature against the information in the class. Use
        :meth:`_verify_issuer` for full verification.

        :param Certificate issuer: The Certificate to verify
        :param VerificationContext context: The context for verification
        :raises SignerInfoVerificationError: If the issuer signature is invalid
        """

        try:
            issuer.verify_signature(
                self.encrypted_digest,
                self._encoded_authenticated_attributes,
                self.digest_algorithm,
                allow_legacy=context.allow_legacy,
            )
        except VerificationError as e:
            raise SignerInfoVerificationError(
                f"Could not verify {issuer} as the signer of the authenticated"
                f" attributes in {type(self).__name__}: {e}"
            )

    def _verify_issuer(
        self,
        issuer: Certificate,
        context: VerificationContext,
        signing_time: datetime.datetime | None = None,
    ) -> list[Certificate]:
        """Verifies whether the given issuer is valid for this :class:`SignerInfo`,
        and valid in the given context. Similar to :meth:`Certificate._verify_issuer`.

        It adds the ``signing_time`` to the context if necessary.
        """

        # _verify_issuer_signature may fail when it is not a valid issuer for
        # this SignedInfo
        self._verify_issuer_signature(issuer, context)

        if signing_time is not None:
            context.timestamp = signing_time
        return context.verify(issuer)

    def _build_chain(
        self,
        context: VerificationContext,
        signing_time: datetime.datetime | None = None,
    ) -> Iterable[list[Certificate]]:
        """Given a context, builds a chain up to a trusted certificate. This is a
        generator function, generating all valid chains.

        This method will call :meth:`VerificationContext.verify` for all possible
        candidates.

        :param VerificationContext context: The context for building the chain. Most
            importantly, contains all certificates to build the chain from, but also
            their properties are relevant.
        :param signing_time: The time to be used as timestamp when creating the chain
        :return: Iterable of all of the valid chains from this SignedInfo up to and
            including a trusted anchor. Note that this may be an empty iteration if no
            candidate parent certificate was found.
        :rtype: Iterable[Iterable[Certificate]]
        :raises AuthenticodeVerificationError: When :meth:`_verify_issuer` fails or
            any of the underlying calls to :meth:`VerificationContext.verify` fails.
            See the semantics of :meth:`VerificationContext.verify` for when that may
            happen. If any error occurs, it is silently swallowed unless no valid chain
            is found. In that case the first error that occurred is raised. If no error
            occurs, no error is raised.
        """

        # this loop was designed in the same way that Certificate._build_chain was built
        # first_error is None until the first iteration. When it becomes False, we do
        # not need to raise anything.
        first_error: VerificationError | None | Literal[False] = None
        for issuer in context.find_certificates(
            issuer=self.issuer, serial_number=self.serial_number
        ):
            try:
                # may fail when anywhere up its chain an error occurs
                yield self._verify_issuer(issuer, context, signing_time)
            except VerificationError as e:  # noqa: PERF203
                if first_error is None:
                    first_error = e
            else:
                first_error = False

        if first_error:
            raise first_error

    def verify(
        self,
        context: VerificationContext,
        signing_time: datetime.datetime | None = None,
    ) -> Iterable[list[Certificate]]:
        """Verifies that this :class:`SignerInfo` verifies up to a chain with the root
        of a trusted certificate.

        :param VerificationContext context: The context for verifying the SignerInfo.
        :param signing_time: The time to be used as timestamp when creating the chain
        :return: A list of valid certificate chains for this SignerInfo.
        :rtype: Iterable[Iterable[Certificate]]
        :raises AuthenticodeVerificationError: When the SignerInfo could not be
            verified.
        """

        chains = list(self._build_chain(context, signing_time))

        if not chains:
            raise SignerInfoVerificationError(
                "No valid certificate chain found to a trust anchor from"
                f" {type(self).__name__}"
            )

        return chains

    def potential_chains(
        self, context: VerificationContext
    ) -> Iterable[list[Certificate]]:
        """Retrieves all potential chains from this SignerInfo instance.

        :param VerificationContext context: The context
        :return: A list of potential certificate chains for this SignerInfo.
        :rtype: Iterable[Iterable[Certificate]]
        """

        for certificate in context.find_certificates(
            issuer=self.issuer, serial_number=self.serial_number
        ):
            yield from context.potential_chains(certificate)


class CounterSignerInfo(SignerInfo):
    """The class CounterSignerInfo is a subclass of :class:`SignerInfo`. It is used as
    the SignerInfo of a SignerInfo, containing the timestamp the SignerInfo was created
    on. This normally works by sending the digest of the SignerInfo to an external
    trusted service, that will include a signed time in its response.
    """

    _required_authenticated_attributes = (
        "content_type",
        "signing_time",
        "message_digest",
    )



# This is a derivative, modified, work from the verify-sigs project.
# Please refer to the LICENSE file in the distribution for more
# information. Original filename: asn1/time_test.py
#
# Parts of this file are licensed as follows:
#
# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from signify.authenticode import TRUSTED_CERTIFICATE_STORE


class RDNSequenceTest(unittest.TestCase):
    def test_to_string(self):
        certificate = TRUSTED_CERTIFICATE_STORE.find_certificate(
            sha256_fingerprint=(
                "DF545BF919A2439C36983B54CDFC903DFA4F37D3996D8D84B4C31EEC6F3C163E"
            )
        )

        self.assertEqual(
            certificate.issuer.dn,
            "CN=Microsoft Root Certificate Authority 2010, O=Microsoft Corporation, "
            "L=Redmond, ST=Washington, C=US",
        )

    def test_to_string_with_commas(self):
        certificate = TRUSTED_CERTIFICATE_STORE.find_certificate(
            sha256_fingerprint=(
                "5B789987F3C4055B8700941B33783A5F16E0CFF937EA32011FE04779F7635308"
            )
        )

        self.assertEqual(
            certificate.issuer.dn,
            r"OU=NO LIABILITY ACCEPTED\, (c)97 VeriSign\, Inc., OU=VeriSign Time"
            r" Stamping Service Root, "
            r"OU=VeriSign\, Inc., O=VeriSign Trust Network",
        )

    def test_get_components(self):
        certificate = TRUSTED_CERTIFICATE_STORE.find_certificate(
            sha256_fingerprint=(
                "5B789987F3C4055B8700941B33783A5F16E0CFF937EA32011FE04779F7635308"
            )
        )

        result = list(certificate.issuer.get_components("OU"))
        self.assertEqual(
            result,
            [
                "NO LIABILITY ACCEPTED, (c)97 VeriSign, Inc.",
                "VeriSign Time Stamping Service Root",
                "VeriSign, Inc.",
            ],
        )
        self.assertEqual(list(certificate.issuer.get_components("CN")), [])

    def test_get_components_none(self):
        certificate = TRUSTED_CERTIFICATE_STORE.find_certificate(
            sha256_fingerprint=(
                "5B789987F3C4055B8700941B33783A5F16E0CFF937EA32011FE04779F7635308"
            )
        )

        result = certificate.issuer.rdns
        self.assertEqual(
            result,
            (
                ("OU", "NO LIABILITY ACCEPTED, (c)97 VeriSign, Inc."),
                ("OU", "VeriSign Time Stamping Service Root"),
                ("OU", "VeriSign, Inc."),
                ("O", "VeriSign Trust Network"),
            ),
        )



from .certificates import Certificate, CertificateName
from .context import CertificateStore, FileSystemCertificateStore, VerificationContext

__all__ = [
    "Certificate",
    "CertificateName",
    "CertificateStore",
    "FileSystemCertificateStore",
    "VerificationContext",
]



import datetime
import pathlib
import unittest

from signify.authenticode import (
    TRUSTED_CERTIFICATE_STORE,
    TRUSTED_CERTIFICATE_STORE_NO_CTL,
)
from signify.authenticode.signed_pe import SignedPEFile
from signify.exceptions import VerificationError
from signify.x509.certificates import Certificate
from signify.x509.context import FileSystemCertificateStore, VerificationContext

root_dir = pathlib.Path(__file__).parent


class TrustedStoreTestCase(unittest.TestCase):
    def test_amount_of_certificates(self):
        self.assertGreaterEqual(len(TRUSTED_CERTIFICATE_STORE), 40)


class ContextTestCase(unittest.TestCase):
    def test_potential_chains(self):
        with open(
            str(
                root_dir
                / "test_data"
                / "19e818d0da361c4feedd456fca63d68d4b024fbbd3d9265f606076c7ee72e8f8.ViR"
            ),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            for signed_data in pefile.signed_datas:
                context = VerificationContext(
                    TRUSTED_CERTIFICATE_STORE_NO_CTL, signed_data.certificates
                )
                potential_chains = list(
                    signed_data.signer_info.potential_chains(context)
                )
                self.assertEqual(len(potential_chains), 2)
                # for chain in potential_chains:
                #    print("xxxx")
                #    for cert in chain:
                #        print(cert)


class ValidationTestCase(unittest.TestCase):
    @unittest.skipIf(
        datetime.datetime.now() > datetime.datetime(2022, 10, 27),
        "revoked certificate expired",
    )
    def test_revoked_certificate(self):
        root = FileSystemCertificateStore(
            root_dir / "certs" / "digicert-global-root-ca.pem", trusted=True
        )
        intermediate = FileSystemCertificateStore(
            root_dir / "certs" / "rapidssl-tls-2020.pem"
        )
        with open(str(root_dir / "certs" / "revoked.badssl.com.pem"), "rb") as f:
            cert = Certificate.from_pem(f.read())

        # check that when we do not verify the CRL it does not fail
        context = VerificationContext(root, intermediate)
        context.verify(cert)

        context = VerificationContext(
            root, intermediate, allow_fetching=True, revocation_mode="hard-fail"
        )
        with self.assertRaises(VerificationError):
            context.verify(cert)



from typing import Any, cast

__version__ = "0.7.0"


def _print_type(t: Any) -> str:
    if t is None:
        return ""
    elif isinstance(t, tuple):
        return ".".join(map(str, t))
    elif callable(t) and hasattr(t(), "name"):
        return cast(str, t().name)  # used by hashlib
    elif hasattr(t, "__name__"):
        return cast(str, t.__name__)
    else:
        return type(t).__name__



#!/usr/bin/env python

# This is a derivative, modified, work from the verify-sigs project.
# Please refer to the LICENSE file in the distribution for more
# information. Original filename: auth_data.py
#
# Parts of this file are licensed as follows:
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module effectively implements Microsoft's documentation on
`<http://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/Authenticode_PE.docx>`_.
"""

from __future__ import annotations

import datetime
import enum
import logging
import pathlib
from typing import Any, Callable, Iterable, Sequence

import mscerts
from asn1crypto import cms, tsp
from typing_extensions import Literal, ParamSpec

from signify import asn1
from signify._typing import HashFunction
from signify.asn1 import spc
from signify.asn1.hashing import _get_digest_algorithm
from signify.asn1.helpers import accuracy_to_python
from signify.authenticode import signed_pe
from signify.authenticode.authroot import CertificateTrustList
from signify.exceptions import (
    AuthenticodeCounterSignerError,
    AuthenticodeInconsistentDigestAlgorithmError,
    AuthenticodeInvalidDigestError,
    AuthenticodeNotSignedError,
    AuthenticodeParseError,
    CertificateVerificationError,
    ParseError,
    SignedPEParseError,
    VerificationError,
)
from signify.pkcs7 import CounterSignerInfo, SignedData, SignerInfo
from signify.x509 import (
    Certificate,
    CertificateName,
    CertificateStore,
    FileSystemCertificateStore,
    VerificationContext,
)

logger = logging.getLogger(__name__)

CERTIFICATE_LOCATION = pathlib.Path(mscerts.where(stl=False))
TRUSTED_CERTIFICATE_STORE_NO_CTL = FileSystemCertificateStore(
    location=CERTIFICATE_LOCATION, trusted=True
)
TRUSTED_CERTIFICATE_STORE = FileSystemCertificateStore(
    location=CERTIFICATE_LOCATION,
    trusted=True,
    ctl=CertificateTrustList.from_stl_file(),
)


_P = ParamSpec("_P")


class AuthenticodeVerificationResult(enum.Enum):
    """This represents the result of an Authenticode verification. If everything is OK,
    it will equal to ``AuthenticodeVerificationResult.OK``, otherwise one of the
    other enum items will be returned. Remember that onl the first exception is
    processed - there may be more wrong.
    """

    OK = enum.auto()
    """The signature is valid."""
    NOT_SIGNED = enum.auto()
    """The provided PE file is not signed."""
    PARSE_ERROR = enum.auto()
    """The Authenticode signature could not be parsed."""
    VERIFY_ERROR = enum.auto()
    """The Authenticode signature could not be verified. This is a more generic error
    than other possible statuses and is used as a catch-all.
    """
    UNKNOWN_ERROR = enum.auto()
    """An unknown error occurred during parsing or verifying."""
    CERTIFICATE_ERROR = enum.auto()
    """An error occurred during the processing of a certificate (e.g. during chain
    building), or when verifying the certificate's signature.
    """
    INCONSISTENT_DIGEST_ALGORITHM = enum.auto()
    """A highly specific error raised when different digest algorithms are used in
    SignedData, SpcInfo or SignerInfo.
    """
    INVALID_DIGEST = enum.auto()
    """The verified digest does not match the calculated digest of the file. This is a
    tell-tale sign that the file may have been tampered with.
    """
    COUNTERSIGNER_ERROR = enum.auto()
    """Something went wrong when verifying the countersignature."""

    @classmethod
    def call(
        cls, function: Callable[_P, Any], *args: _P.args, **kwargs: _P.kwargs
    ) -> tuple[AuthenticodeVerificationResult, Exception | None]:
        try:
            function(*args, **kwargs)
        except (SignedPEParseError, AuthenticodeNotSignedError) as exc:
            return cls.NOT_SIGNED, exc
        except AuthenticodeInconsistentDigestAlgorithmError as exc:
            return cls.INCONSISTENT_DIGEST_ALGORITHM, exc
        except AuthenticodeInvalidDigestError as exc:
            return cls.INVALID_DIGEST, exc
        except AuthenticodeCounterSignerError as exc:
            return cls.COUNTERSIGNER_ERROR, exc
        except CertificateVerificationError as exc:
            return cls.CERTIFICATE_ERROR, exc
        except ParseError as exc:
            return cls.PARSE_ERROR, exc
        except VerificationError as exc:
            return cls.VERIFY_ERROR, exc
        except Exception as exc:
            return cls.UNKNOWN_ERROR, exc
        else:
            return cls.OK, None


class AuthenticodeCounterSignerInfo(CounterSignerInfo):
    """Subclass of :class:`CounterSignerInfo` that is used to contain the
    countersignerinfo for Authenticode.
    """


class AuthenticodeSignerInfo(SignerInfo):
    """Subclass of :class:`SignerInfo` that is used by the verification of Authenticode.
    Note that this will contain the same attributes as :class:`SignerInfo`, and
    additionally the following:

    .. attribute:: program_name
                   more_info
                   publisher_info

       This information is extracted from the SpcSpOpusInfo authenticated attribute,
       containing the program's name and an URL with more information. The
       publisher_info is almost never set, but is defined in the ASN.1 structure.

    .. attribute:: statement_types

       Defines the key purpose of the signer. This is ignored by the verification.

    .. attribute:: nested_signed_datas

       It is possible for Authenticode SignerInfo objects to contain nested
       :class:`signify.pkcs7.SignedData` objects. This is  similar to including
       multiple SignedData structures in the :class:`signify.authenticode.SignedPEFile`.
       This field  is extracted from  the unauthenticated attributes.

    The :attr:`countersigner` attribute can hold the same as in the normal
    :class:`SignerInfo`, but may also contain a :class:`RFC3161SignedData` class:

    .. attribute:: countersigner

       Authenticode may use a different countersigning mechanism, rather than using a
       nested :class:`AuthenticodeCounterSignerInfo`, it  may use a nested RFC-3161
       response, which is a nested :class:`signify.pkcs7.SignedData` structure
       (of type :class:`RFC3161SignedData`). This is also assigned to the countersigner
       attribute if this is available.


    """

    program_name: str | None
    more_info: str | None
    publisher_info: str | None
    nested_signed_datas: list[AuthenticodeSignedData]

    parent: AuthenticodeSignedData
    # allow other countersigner as well
    countersigner: (  # type: ignore[assignment]
        AuthenticodeCounterSignerInfo | RFC3161SignedData | None
    )

    _countersigner_class = AuthenticodeCounterSignerInfo
    _expected_content_type = "microsoft_spc_indirect_data_content"

    def _parse(self) -> None:
        super()._parse()

        # - Retrieve statement types
        self.statement_types = None
        if "microsoft_spc_statement_type" in self.authenticated_attributes:
            if len(self.authenticated_attributes["microsoft_spc_statement_type"]) != 1:
                raise AuthenticodeParseError(
                    "Only one SpcStatementType expected in"
                    " SignerInfo.authenticatedAttributes"
                )
            self.statement_types = self.authenticated_attributes[
                "microsoft_spc_statement_type"
            ][0].native

        # - Retrieve object from SpcSpOpusInfo from the authenticated attributes
        # (for normal signer)
        self.program_name = self.more_info = self.publisher_info = None
        if "microsoft_spc_sp_opus_info" in self.authenticated_attributes:
            if len(self.authenticated_attributes["microsoft_spc_sp_opus_info"]) != 1:
                raise AuthenticodeParseError(
                    "Only one SpcSpOpusInfo expected in"
                    " SignerInfo.authenticatedAttributes"
                )

            self.program_name = self.authenticated_attributes[
                "microsoft_spc_sp_opus_info"
            ][0]["program_name"].native
            self.more_info = self.authenticated_attributes[
                "microsoft_spc_sp_opus_info"
            ][0]["more_info"].native
            self.publisher_info = self.authenticated_attributes[
                "microsoft_spc_sp_opus_info"
            ][0]["publisher_info"].native

        # - Authenticode can use nested signatures through OID 1.3.6.1.4.1.311.2.4.1
        self.nested_signed_datas = []
        if "microsoft_nested_signature" in self.unauthenticated_attributes:
            for sig_data in self.unauthenticated_attributes[
                "microsoft_nested_signature"
            ]:
                content_type = sig_data["content_type"].native
                if content_type != "signed_data":
                    raise AuthenticodeParseError(
                        "Nested signature is not a SignedData structure"
                    )
                self.nested_signed_datas.append(
                    AuthenticodeSignedData(
                        sig_data["content"], pefile=self.parent.pefile
                    )
                )

        # - Authenticode can be signed using a RFC-3161 timestamp, so we discover this
        # possibility here
        if (
            "counter_signature" in self.unauthenticated_attributes
            and "microsoft_time_stamp_token" in self.unauthenticated_attributes
        ):
            raise AuthenticodeParseError(
                "Countersignature and RFC-3161 timestamp present in"
                " SignerInfo.unauthenticatedAttributes"
            )

        if "microsoft_time_stamp_token" in self.unauthenticated_attributes:
            if len(self.unauthenticated_attributes["microsoft_time_stamp_token"]) != 1:
                raise AuthenticodeParseError(
                    "Only one RFC-3161 timestamp expected in"
                    " SignerInfo.unauthenticatedAttributes"
                )

            ts_data = self.unauthenticated_attributes["microsoft_time_stamp_token"][0]
            if ts_data["content_type"].native != "signed_data":
                raise AuthenticodeParseError(
                    "RFC-3161 Timestamp does not contain SignedData structure"
                )

            self.countersigner = RFC3161SignedData(ts_data["content"])

    def _verify_issuer(
        self,
        issuer: Certificate,
        context: VerificationContext,
        signing_time: datetime.datetime | None = None,
    ) -> list[Certificate]:
        """Check whether the lifetime signing EKU is set. if that is the case, we can
        only use the timestamp for revocation checking, not for extending the lifetime
        of the signature. Revocation checking currently does not work.
        """
        if "microsoft_lifetime_signing" in issuer.extensions.get(
            "extended_key_usage", []
        ):
            signing_time = None
        return super()._verify_issuer(issuer, context, signing_time)


class SpcInfo:
    """The Authenticode's SpcIndirectDataContent information, and their children. This
    is expected to be part of the content of the SignedData structure in Authenticode.

    Note that this structure is completely flattened out from this ASN.1 spec::

        SpcIndirectDataContent ::= SEQUENCE {
            data SpcAttributeTypeAndOptionalValue,
            messageDigest  DigestInfo
        }
        SpcAttributeTypeAndOptionalValue ::= SEQUENCE {
            type ObjectID,
            value [0] EXPLICIT ANY OPTIONAL
        }
        DigestInfo ::= SEQUENCE {
            digestAlgorithm  AlgorithmIdentifier,
            digest OCTETSTRING
        }
        AlgorithmIdentifier ::= SEQUENCE {
            algorithm ObjectID,
            parameters [0] EXPLICIT ANY OPTIONAL
        }

    .. attribute:: data

       The underlying ASN.1 data object

    .. attribute:: content_type

       The contenttype string

    .. attribute:: image_data

       The image data object embedded in the ASN.1 object.

    .. attribute:: image_flags

       The flags used for signing. These flags are ignored during verification.

    .. attribute:: image_publisher

       Obsolete software publisher field (i.e. ``SpcPeImageData.file``). Should now
       contain ``<<<Obsolete>>>``, although this value does not affect verification.

    .. attribute:: digest_algorithm
    .. attribute:: digest


    """

    data: spc.SpcIndirectDataContent

    content_type: str
    image_data: spc.SpcPeImageData
    image_flags: set[str]
    image_publisher: str
    digest_algorithm: HashFunction
    digest: bytes

    def __init__(self, data: spc.SpcIndirectDataContent):
        self.data = data
        self._parse()

    def _parse(self) -> None:
        # The data attribute
        self.content_type = self.data["data"]["type"].native

        if self.content_type != "microsoft_spc_pe_image_data":
            raise AuthenticodeParseError("SpcInfo does not contain SpcPeImageData")

        self.image_data = self.data["data"]["value"]
        self.image_flags = self.image_data["flags"].native
        self.image_publisher = self.image_data["file"].native

        self.digest_algorithm = _get_digest_algorithm(
            self.data["message_digest"]["digest_algorithm"],
            location="SpcIndirectDataContent.digestAlgorithm",
        )
        self.digest = self.data["message_digest"]["digest"].native


class AuthenticodeSignedData(SignedData):
    """The :class:`signify.pkcs7.SignedData` structure for Authenticode. It holds the
    same information as its superclass, with additionally the :class:`SpcInfo`:

    .. attribute:: spc_info

       The parsed :attr:`content` of this :class:`SignedData` object, being a
       SpcIndirectDataContent object.

    """

    pefile: signed_pe.SignedPEFile | None
    spc_info: SpcInfo
    signer_infos: Sequence[AuthenticodeSignerInfo]
    signer_info: AuthenticodeSignerInfo

    content: asn1.spc.SpcIndirectDataContent
    _expected_content_type = "microsoft_spc_indirect_data_content"
    _signerinfo_class = AuthenticodeSignerInfo

    def __init__(
        self,
        data: cms.SignedData,
        pefile: signed_pe.SignedPEFile | None = None,
    ):
        """
        :param asn1.pkcs7.SignedData data: The ASN.1 structure of the SignedData object
        :param pefile: The related PEFile.
        """
        self.pefile = pefile
        super().__init__(data)

    def _parse(self) -> None:
        super()._parse()
        self.spc_info = SpcInfo(self.content)

        # signerInfos
        if len(self.signer_infos) != 1:
            raise AuthenticodeParseError(
                "SignedData.signerInfos must contain exactly 1 signer,"
                f" not {len(self.signer_infos)}"
            )

        self.signer_info = self.signer_infos[0]

        # CRLs
        if self.data["crls"]:
            raise AuthenticodeParseError(
                "SignedData.crls is present, but that is unexpected."
            )

    def verify(
        self,
        *,
        expected_hash: bytes | None = None,
        verification_context: VerificationContext | None = None,
        cs_verification_context: VerificationContext | None = None,
        trusted_certificate_store: CertificateStore = TRUSTED_CERTIFICATE_STORE,
        verification_context_kwargs: dict[str, Any] | None = None,
        countersignature_mode: Literal["strict", "permit", "ignore"] = "strict",
    ) -> Iterable[list[Certificate]]:
        """Verifies the SignedData structure:

        * Verifies that the digest algorithms match across the structure
          (:class:`SpcInfo`, :class:`AuthenticodeSignedData` and
          :class:`AuthenticodeSignerInfo` must have the same)
        * Ensures that the hash in :attr:`SpcInfo.digest` matches the expected hash.
          If no expected hash is provided to this function, it is calculated using
          the :class:`Fingerprinter` obtained from the :class:`SignedPEFile` object.
        * Verifies that the :class:`SpcInfo`, when hashed, is the same as the value in
          :attr:`SignerInfo.message_digest`
        * In the case of a countersigner, calls :meth:`check_message_digest` on the
          countersigner to verify that the hashed value of
          :attr:`AuthenticodeSignerInfo.encrypted_digest` is contained in the
          countersigner.
        * Verifies the chain of the countersigner up to a trusted root, see
          :meth:`SignerInfo.verify` and :meth:`RFC3161SignedData.verify`
        * Verifies the chain of the signer up to a trusted root, see
          :meth:`SignerInfo.verify`

        In the case of a countersigner, the verification is performed using the
        timestamp of the :class:`CounterSignerInfo`, otherwise now is assumed. If there
        is no countersigner, you can override this by specifying a different timestamp
        in the :class:`VerificationContext`. Note that you cannot set a timestamp when
        checking against the CRL; this is not permitted by the underlying library. If
        you need to do this, you must therefore set countersignature_mode to ``ignore``.

        :param bytes expected_hash: The expected hash digest of the
            :class:`SignedPEFile`.
        :param VerificationContext verification_context: The VerificationContext for
            verifying the chain of the :class:`SignerInfo`. The timestamp is overridden
            in the case of a countersigner. Default stores are TRUSTED_CERTIFICATE_STORE
            and the certificates of this :class:`SignedData` object. EKU is code_signing
        :param VerificationContext cs_verification_context: The VerificationContext for
            verifying the chain of the :class:`CounterSignerInfo`. The timestamp is
            overridden in the case of a countersigner. Default stores are
            TRUSTED_CERTIFICATE_STORE and the certificates of this :class:`SignedData`
            object. EKU is time_stamping
        :param CertificateStore trusted_certificate_store: A :class:`CertificateStore`
            object that contains a list of trusted certificates to be used when
            :const:`None` is passed to either ``verification_context`` or
            ``cs_verification_context`` and a :class:`VerificationContext` is created.
        :param dict verification_context_kwargs: If provided, keyword arguments that
            are passed to the instantiation of :class:`VerificationContext` s created
            in this function. Used for e.g. providing a timestamp.
        :param str countersignature_mode: Changes how countersignatures are handled.
            Defaults to 'strict', which means that errors in the countersignature
            result in verification failure. If set to 'permit', the countersignature is
            checked, but when it errors, it is verified as if the countersignature was
            never set. When set to 'ignore', countersignatures are never checked.
        :raises AuthenticodeVerificationError: when the verification failed
        :return: A list of valid certificate chains for this SignedData.
        """

        if verification_context_kwargs is None:
            verification_context_kwargs = {}
        if verification_context is None:
            verification_context = VerificationContext(
                trusted_certificate_store,
                self.certificates,
                extended_key_usages=["code_signing"],
                **verification_context_kwargs,
            )

        if (
            cs_verification_context is None
            and self.signer_info.countersigner
            and countersignature_mode != "ignore"
        ):
            cs_verification_context = VerificationContext(
                trusted_certificate_store,
                self.certificates,
                extended_key_usages=["time_stamping"],
                **verification_context_kwargs,
            )
            # Add the local certificate store for the countersignature
            # (in the case of RFC3161SignedData)
            if hasattr(self.signer_info.countersigner, "certificates"):
                cs_verification_context.add_store(
                    self.signer_info.countersigner.certificates
                )

        # Check that the digest algorithms match
        if self.digest_algorithm != self.spc_info.digest_algorithm:
            raise AuthenticodeInconsistentDigestAlgorithmError(
                "SignedData.digestAlgorithm must equal SpcInfo.digestAlgorithm"
            )

        if self.digest_algorithm != self.signer_info.digest_algorithm:
            raise AuthenticodeInconsistentDigestAlgorithmError(
                "SignedData.digestAlgorithm must equal SignerInfo.digestAlgorithm"
            )

        # Check that the hashes are correct
        # 1. The hash of the file
        if expected_hash is None:
            assert self.pefile is not None
            fingerprinter = self.pefile.get_fingerprinter()
            fingerprinter.add_authenticode_hashers(self.digest_algorithm)
            expected_hash = fingerprinter.hash()[self.digest_algorithm().name]

        if expected_hash != self.spc_info.digest:
            raise AuthenticodeInvalidDigestError(
                "The expected hash does not match the digest in SpcInfo"
            )

        # 2. The hash of the spc blob
        if self.content_digest != self.signer_info.message_digest:
            raise AuthenticodeInvalidDigestError(
                "The expected hash of the SpcInfo does not match SignerInfo"
            )

        # Can't check authAttr hash against encrypted hash, done implicitly in
        # M2's pubkey.verify.

        signing_time = None
        if self.signer_info.countersigner and countersignature_mode != "ignore":
            assert cs_verification_context is not None

            try:
                # 3. Check the countersigner hash.
                # Make sure to use the same digest_algorithm that the countersigner used
                if not self.signer_info.countersigner.check_message_digest(
                    self.signer_info.encrypted_digest
                ):
                    raise AuthenticodeCounterSignerError(
                        "The expected hash of the encryptedDigest does not match"
                        " countersigner's SignerInfo"
                    )

                cs_verification_context.timestamp = (
                    self.signer_info.countersigner.signing_time
                )

                # We could be calling SignerInfo.verify or RFC3161SignedData.verify
                # here, but those have identical signatures. Note that
                # RFC3161SignedData accepts a trusted_certificate_store argument, but
                # we pass in an explicit context anyway
                self.signer_info.countersigner.verify(cs_verification_context)
            except Exception as e:
                if countersignature_mode != "strict":
                    pass
                else:
                    raise AuthenticodeCounterSignerError(
                        f"An error occurred while validating the countersignature: {e}"
                    )
            else:
                # If no errors occur, we should be fine setting the timestamp to the
                # countersignature's timestamp
                signing_time = self.signer_info.countersigner.signing_time

        return self.signer_info.verify(verification_context, signing_time)

    def explain_verify(
        self, *args: Any, **kwargs: Any
    ) -> tuple[AuthenticodeVerificationResult, Exception | None]:
        """This will return a value indicating the signature status of this object.
        This will not raise an error when the verification fails, but rather indicate
        this through the resulting enum

        :rtype: Tuple[AuthenticodeVerificationResult, Exception]
        :return: The verification result, and the exception containing more details
            (if available or None)
        """

        return AuthenticodeVerificationResult.call(self.verify, *args, **kwargs)


class RFC3161SignerInfo(SignerInfo):
    """Subclass of SignerInfo that is used to contain the signerinfo for the
    RFC3161SignedData option.
    """

    _expected_content_type = "tst_info"
    _countersigner_class = None  # prevent countersigners in here


class TSTInfo:
    """This is an implementation of the TSTInfo class as defined by RFC3161, used as
    content for a SignedData structure. The following properties are available:

    .. attribute:: data

       The underlying ASN.1 data object

    .. attribute:: policy

    .. attribute:: hash_algorithm

       The hash algorithm of the message imprint.

    .. attribute:: message_digest

       The hashed message

    .. attribute:: serial_number

       The serial number of this signature

    .. attribute:: signing_time

       The time this signature was generated

    .. attribute:: signing_time_accuracy

       The accuracy of the above time

    .. attribute:: signing_authority

       The authority generating this signature

    """

    policy: str
    hash_algorithm: HashFunction
    message_digest: bytes
    serial_number: int
    signing_time: datetime.datetime
    signing_time_accuracy: datetime.timedelta
    signing_authority: CertificateName

    def __init__(self, data: tsp.TSTInfo):
        """

        :param data: The ASN.1 structure of the TSTInfo object
        """
        self.data = data
        self._parse()

    def _parse(self) -> None:
        if self.data["version"].native != "v1":
            raise AuthenticodeParseError(
                f"TSTInfo.version must be v1, not {self.data['version'].native}"
            )

        self.policy = self.data["policy"].native
        self.hash_algorithm = _get_digest_algorithm(
            self.data["message_imprint"]["hash_algorithm"],
            location="TSTInfo.messageImprint.hashAlgorithm",
        )
        self.message_digest = self.data["message_imprint"]["hashed_message"].native
        self.serial_number = self.data["serial_number"].native
        self.signing_time = self.data["gen_time"].native
        self.signing_time_accuracy = accuracy_to_python(self.data["accuracy"])
        self.signing_authority = CertificateName(self.data["tsa"])


class RFC3161SignedData(SignedData):
    """Some samples have shown to include a RFC-3161 countersignature in the
    unauthenticated attributes (as OID 1.3.6.1.4.1.311.3.3.1, which is in the Microsoft
    private namespace). This attribute contains its own signed data structure.

    This is a subclass of :class:`signify.pkcs7.SignedData`, containing a RFC3161
    TSTInfo in its content field.

    .. attribute:: tst_info
       :type: TSTInfo

       Contains the :class:`TSTInfo` class for this SignedData.
    """

    content: tsp.TSTInfo
    _expected_content_type = "tst_info"
    _signerinfo_class = RFC3161SignerInfo

    def _parse(self) -> None:
        super()._parse()

        # Get the tst_info
        self.tst_info = TSTInfo(self.content)

        # signerInfos
        if len(self.signer_infos) != 1:
            raise AuthenticodeParseError(
                "RFC3161 SignedData.signerInfos must contain exactly 1 signer,"
                f" not {len(self.signer_infos)}"
            )

        self.signer_info = self.signer_infos[0]

    @property
    def signing_time(self) -> datetime.datetime:
        """Transparent attribute to ensure that the signing_time attribute is
        consistently available.
        """
        return self.tst_info.signing_time

    def check_message_digest(self, data: bytes) -> bool:
        """Given the data, returns whether the hash_algorithm and message_digest match
        the data provided.
        """

        auth_attr_hasher = self.tst_info.hash_algorithm()
        auth_attr_hasher.update(data)
        return auth_attr_hasher.digest() == self.tst_info.message_digest

    def verify(
        self,
        context: VerificationContext | None = None,
        *,
        trusted_certificate_store: CertificateStore = TRUSTED_CERTIFICATE_STORE,
    ) -> Iterable[Iterable[Certificate]]:
        """Verifies the RFC3161 SignedData object. The context that is passed in must
        account for the certificate store of this object, or be left None.

        The object is verified by verifying that the hash of the :class:`TSTInfo`
        matches the :attr:`SignerInfo.message_digest` value. The remainder of the
        validation is done by calling :meth:`SignerInfo.verify`
        """

        # We should ensure that the hash in the SignerInfo matches the hash of the
        # content. This is similar to the normal verification process, where the
        # SpcInfo is verified. Note that the mapping between the RFC3161 SignedData
        # object is ensured by the verifier in SignedData
        if self.content_digest != self.signer_info.message_digest:
            raise AuthenticodeCounterSignerError(
                "The expected hash of the TstInfo does not match SignerInfo"
            )

        if context is None:
            context = VerificationContext(
                trusted_certificate_store,
                self.certificates,
                extended_key_usages=["time_stamping"],
            )

        # The context is set correctly by the 'verify' function, including the current
        # certificate store
        return self.signer_info.verify(context)


if __name__ == "__main__":
    print(
        "This is a list of all certificates in the Authenticode trust store, ordered by"
        " expiration date"
    )
    for i, certificate in enumerate(
        sorted(TRUSTED_CERTIFICATE_STORE, key=lambda x: x.valid_to), start=1
    ):
        print(i, certificate.valid_to, certificate)



class SignifyError(Exception):
    """Base class for all errors generated by Signify"""


class ParseError(SignifyError):
    pass


class SignedPEParseError(ParseError):
    pass


class SignerInfoParseError(ParseError):
    pass


class AuthenticodeParseError(ParseError):
    """Raised when any exception regarding parsing Authenticode structures occurs."""


class VerificationError(SignifyError):
    pass


class CertificateVerificationError(VerificationError):
    pass


class CertificateNotTrustedVerificationError(CertificateVerificationError):
    pass


class SignerInfoVerificationError(VerificationError):
    pass


class AuthenticodeVerificationError(VerificationError):
    """Raised when any exception regarding verifying Authenticode structures occurs."""


class AuthenticodeNotSignedError(AuthenticodeVerificationError):
    pass


class AuthenticodeInconsistentDigestAlgorithmError(AuthenticodeVerificationError):
    pass


class AuthenticodeInvalidDigestError(AuthenticodeVerificationError):
    pass


class AuthenticodeCounterSignerError(AuthenticodeVerificationError):
    pass


class CertificateTrustListParseError(SignifyError):
    pass


class CTLCertificateVerificationError(CertificateVerificationError):
    pass



# This is a derivative, modified, work from the verify-sigs project.
# Please refer to the LICENSE file in the distribution for more
# information. Original filename: auth_data_test.py
#
# Parts of this file are licensed as follows:
#
# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import binascii
import datetime
import hashlib
import io
import pathlib
import unittest

from signify.authenticode import (
    CERTIFICATE_LOCATION,
    TRUSTED_CERTIFICATE_STORE,
    TRUSTED_CERTIFICATE_STORE_NO_CTL,
)
from signify.authenticode.signed_pe import SignedPEFile
from signify.exceptions import (
    AuthenticodeNotSignedError,
    AuthenticodeVerificationError,
    SignedPEParseError,
    VerificationError,
)
from signify.fingerprinter import AuthenticodeFingerprinter
from signify.x509 import Certificate
from signify.x509.context import (
    CertificateStore,
    FileSystemCertificateStore,
    VerificationContext,
)

root_dir = pathlib.Path(__file__).parent
trusted_certificate_store = FileSystemCertificateStore(
    location=CERTIFICATE_LOCATION, trusted=True
)


class AuthenticodeParserTestCase(unittest.TestCase):
    def test_software_update(self):
        with open(str(root_dir / "test_data" / "SoftwareUpdate.exe"), "rb") as f:
            fingerprinter = AuthenticodeFingerprinter(f)
            fingerprinter.add_authenticode_hashers(hashlib.sha1)
            hashes = fingerprinter.hash()

            # Sanity check that the authenticode hash is still correct
            self.assertEqual(
                binascii.hexlify(hashes["sha1"]).decode("ascii"),
                "978b90ace99c764841d2dd17d278fac4149962a3",
            )

            pefile = SignedPEFile(f)

            # This should not raise any errors.
            signed_datas = list(pefile.signed_datas)
            # There may be multiple of these, if the windows binary was signed multiple
            # times, e.g. by different entities. Each of them adds a complete SignedData
            # blob to the binary. For our sample, there is only one blob.
            self.assertEqual(len(signed_datas), 1)
            signed_data = signed_datas[0]

            signed_data.verify()

            # should work as well
            pefile.verify()

    def test_pciide(self):
        with open(str(root_dir / "test_data" / "pciide.sys"), "rb") as f:
            pefile = SignedPEFile(f)
            signed_datas = list(pefile.signed_datas)
            self.assertEqual(len(signed_datas), 1)
            signed_data = signed_datas[0]
            signed_data.verify()
            pefile.verify()

    def test_modified_pciide_fails(self):
        with open(str(root_dir / "test_data" / "pciide.sys"), "rb") as f:
            data = bytearray(f.read())
        data[1024] = 3
        bt = io.BytesIO(data)
        pefile = SignedPEFile(bt)
        signed_datas = list(pefile.signed_datas)
        self.assertEqual(len(signed_datas), 1)
        self.assertRaises(AuthenticodeVerificationError, signed_datas[0].verify)
        self.assertRaises(AuthenticodeVerificationError, pefile.verify)

    def test_simple(self):
        with open(str(root_dir / "test_data" / "simple"), "rb") as f:
            pefile = SignedPEFile(f)
            self.assertRaises(AuthenticodeNotSignedError, pefile.verify)
            self.assertRaises(
                SignedPEParseError,
                list,
                pefile.iter_signed_datas(ignore_parse_errors=False),
            )

    def test_2A6E(self):
        with open(str(root_dir / "test_data" / "___2A6E.tmp"), "rb") as f:
            pefile = SignedPEFile(f)
            self.assertRaises(VerificationError, pefile.verify)

    def test_0d8c_valid(self):
        with open(
            str(
                root_dir
                / "test_data"
                / "0d8c2bcb575378f6a88d17b5f6ce70e794a264cdc8556c8e812f0b5f9c709198"
            ),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            pefile.verify(trusted_certificate_store=TRUSTED_CERTIFICATE_STORE_NO_CTL)

    def test_provide_hash(self):
        with open(
            str(
                root_dir
                / "test_data"
                / "0d8c2bcb575378f6a88d17b5f6ce70e794a264cdc8556c8e812f0b5f9c709198"
            ),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            with self.assertRaises(VerificationError):
                pefile.verify(
                    trusted_certificate_store=TRUSTED_CERTIFICATE_STORE_NO_CTL,
                    expected_hashes={"sha1": b"asdf"},
                )

    def test_19e8_expired(self):
        """this is an expired sample"""
        with open(
            str(
                root_dir
                / "test_data"
                / "19e818d0da361c4feedd456fca63d68d4b024fbbd3d9265f606076c7ee72e8f8.ViR"
            ),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            self.assertRaises(VerificationError, pefile.verify)

    def test_19e8_valid_within_period(self):
        """test whether the timestamp can be set on expired samples"""
        with open(
            str(
                root_dir
                / "test_data"
                / "19e818d0da361c4feedd456fca63d68d4b024fbbd3d9265f606076c7ee72e8f8.ViR"
            ),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            pefile.verify(
                verification_context_kwargs={
                    "timestamp": datetime.datetime(
                        2013, 1, 1, tzinfo=datetime.timezone.utc
                    )
                }
            )

    def test_sw_reporter(self):
        """Test for SHA256 hashes used in sig"""
        with open(
            str(root_dir / "test_data" / "software_reporter_tool.exe"), "rb"
        ) as f:
            pefile = SignedPEFile(f)
            signed_datas = list(pefile.signed_datas)
            self.assertEqual(len(signed_datas), 1)
            signed_data = signed_datas[0]
            signed_data.verify()
            pefile.verify()

    def test_7z1900_invalid_cve2020_0601(self):
        """This tests against CVE-2020-0601"""
        with open(str(root_dir / "test_data" / "7z1900-x64_signed.exe"), "rb") as f:
            pefile = SignedPEFile(f)
            self.assertRaises(VerificationError, pefile.verify)

    def test_3a7de393a36ca8911cd0842a9a25b058_valid_different_contenttype(self):
        """uses a different contenttype, 1.2.840.113549.1.9.16.1.4 instead of Data"""
        with open(
            str(root_dir / "test_data" / "3a7de393a36ca8911cd0842a9a25b058"), "rb"
        ) as f:
            pefile = SignedPEFile(f)
            pefile.verify()

    def test_solwarwinds_valid_countersignature_rfc3161(self):
        """Solarwinds includes a 1.3.6.1.4.1.311.3.3.1 type countersignature"""
        with open(str(root_dir / "test_data" / "SolarWinds.exe"), "rb") as f:
            pefile = SignedPEFile(f)
            pefile.verify()

    def test_whois_valid_countersignature_rfc3161(self):
        """whois includes a 1.3.6.1.4.1.311.3.3.1 type countersignature"""
        with open(str(root_dir / "test_data" / "whois.exe"), "rb") as f:
            pefile = SignedPEFile(f)
            pefile.verify()

    def test_jameslth_valid_when_revocation_not_checked(self):
        """this certificate is revoked"""
        with open(str(root_dir / "test_data" / "jameslth"), "rb") as f:
            pefile = SignedPEFile(f)
            pefile.verify(
                verification_context_kwargs={
                    "timestamp": datetime.datetime(
                        2021, 1, 1, tzinfo=datetime.timezone.utc
                    )
                }
            )

    def test_jameslth_revoked(self):
        """this certificate is revoked"""
        # TODO: this certificate is now expired, so it will not show up as valid anyway
        with open(str(root_dir / "test_data" / "jameslth"), "rb") as f:
            pefile = SignedPEFile(f)
            with self.assertRaises(VerificationError):
                pefile.verify(
                    verification_context_kwargs={
                        "allow_fetching": True,
                        "revocation_mode": "hard-fail",
                    }
                )

    def test_zonealarm_rfc3161_different_hash_and_digest_algorithms(self):
        """this tests a RFC3161 sample that has distinct hash and digest algorithms"""
        with open(str(root_dir / "test_data" / "zonealarm.exe"), "rb") as f:
            pefile = SignedPEFile(f)
            pefile.verify()

    def test_abnormal_attribute_order(self):
        """this tests a sample that has an abnormal attribute order"""
        with open(
            str(root_dir / "test_data" / "8757bf55-0077-4df5-9807-122a3261ee40"), "rb"
        ) as f:
            pefile = SignedPEFile(f)
            pefile.verify()

    def test_multiple_signatures_all_valid(self):
        """this tests a sample that has two signed datas, that are both valid"""
        with open(str(root_dir / "test_data" / "sigcheck.exe"), "rb") as f:
            pefile = SignedPEFile(f)
            self.assertEqual(len(list(pefile.signed_datas)), 2)

            for mode in ("all", "first", "any", "best"):
                with self.subTest(multi_verify_mode=mode):
                    pefile.verify(
                        trusted_certificate_store=TRUSTED_CERTIFICATE_STORE_NO_CTL,
                        multi_verify_mode=mode,
                    )

    def test_multiple_signatures_invalid_sha1(self):
        """this tests a sample that has an invalid sha-1 hash, but valid sha-256 hash"""
        with open(str(root_dir / "test_data" / "sigcheck_sha1_patched.exe"), "rb") as f:
            pefile = SignedPEFile(f)
            for mode in ("all", "first"):
                with self.subTest(multi_verify_mode=mode), self.assertRaises(
                    VerificationError
                ):
                    pefile.verify(
                        trusted_certificate_store=TRUSTED_CERTIFICATE_STORE_NO_CTL,
                        multi_verify_mode=mode,
                    )
            for mode in ("any", "best"):
                with self.subTest(multi_verify_mode=mode):
                    pefile.verify(
                        trusted_certificate_store=TRUSTED_CERTIFICATE_STORE_NO_CTL,
                        multi_verify_mode=mode,
                    )

    def test_multiple_signatures_invalid_sha256(self):
        """this tests a sample that has an valid sha-1 hash, but invalid sha-256 hash"""
        with open(
            str(root_dir / "test_data" / "sigcheck_sha256_patched.exe"), "rb"
        ) as f:
            pefile = SignedPEFile(f)
            for mode in ("all", "best"):
                with self.subTest(multi_verify_mode=mode), self.assertRaises(
                    VerificationError
                ):
                    pefile.verify(
                        trusted_certificate_store=TRUSTED_CERTIFICATE_STORE_NO_CTL,
                        multi_verify_mode=mode,
                    )
            for mode in ("first", "any"):
                with self.subTest(multi_verify_mode=mode):
                    pefile.verify(
                        trusted_certificate_store=TRUSTED_CERTIFICATE_STORE_NO_CTL,
                        multi_verify_mode=mode,
                    )

    def test_multiple_signatures_all_invalid(self):
        """this tests a sample that has only invalid signautres"""
        with open(
            str(root_dir / "test_data" / "sigcheck_sha256_patched.exe"), "rb"
        ) as f:
            pefile = SignedPEFile(f)
            # we can test for the fact that all signatures are invalid here as well,
            # because the normal CTL will disallow sha1
            for mode in ("all", "best", "first", "any"):
                with self.subTest(multi_verify_mode=mode), self.assertRaises(
                    VerificationError
                ):
                    pefile.verify(multi_verify_mode=mode)

    def test_signeddata_v3(self):
        """this tests a sample that has a v3 SignedData structure"""
        with open(
            str(
                root_dir
                / "test_data"
                / "7dc674b46d51cb42963417a487ef8e88e547e3e0902a1e236ff13c3f1fdd60e4.exe"
            ),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            pefile.verify()

    def test_signerinfo_v0(self):
        """this tests a sample that has a v0 SignerInfo structure"""
        with open(
            str(root_dir / "test_data" / "spotify_sample.exe"),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            pefile.verify()

    def test_lifetime_signing(self):
        """this tests a sample that has a valid countersignature and a lifetime signing
        flag. it is self-signed, so we need to load that one as well
        """
        certificate_store = CertificateStore(
            list(TRUSTED_CERTIFICATE_STORE_NO_CTL)
            + list(
                Certificate.from_pems(
                    (root_dir / "test_data" / "kdbazis.dll.crt").read_bytes()
                )
            ),
            trusted=True,
        )

        with open(
            str(root_dir / "test_data" / "kdbazis.dll"),
            "rb",
        ) as f:
            pefile = SignedPEFile(f)
            # should verify on this date
            pefile.verify(
                trusted_certificate_store=certificate_store,
                verification_context_kwargs={
                    "timestamp": datetime.datetime(
                        2024, 1, 1, tzinfo=datetime.timezone.utc
                    )
                },
            )
            # should not verify on this date
            with self.assertRaises(VerificationError):
                pefile.verify(
                    trusted_certificate_store=certificate_store,
                    verification_context_kwargs={
                        "timestamp": datetime.datetime(
                            2060, 1, 1, tzinfo=datetime.timezone.utc
                        )
                    },
                )


class CertificateTestCase(unittest.TestCase):
    def test_all_trusted_certificates_are_trusted(self):
        context = VerificationContext(trusted_certificate_store)
        # only select 50 to speed up testing
        for certificate in trusted_certificate_store[:50]:
            # Trust depends on the timestamp
            context.timestamp = certificate.valid_to
            chain = certificate.verify(context)
            self.assertListEqual(chain, [certificate])

    def test_no_duplicates_in_default_store(self):
        self.assertEqual(
            len(TRUSTED_CERTIFICATE_STORE), len(set(TRUSTED_CERTIFICATE_STORE))
        )

    def test_trust_fails(self):
        # we get a certificate we currently trust
        for certificate in trusted_certificate_store:
            # we add it to an untrusted store
            store = CertificateStore(trusted=False)
            store.append(certificate)
            # and verify using this store
            context = VerificationContext(store, timestamp=certificate.valid_to)
            self.assertRaises(VerificationError, certificate.verify, context)



#!/usr/bin/env python

# This is a derivative, modified, work from the verify-sigs project.
# Please refer to the LICENSE file in the distribution for more
# information. Original filename: fingerprinter.py
#
# Parts of this file are licensed as follows:
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import binascii
import collections
import hashlib
import logging
import os
import pathlib
import sys
from typing import BinaryIO

from signify._typing import HashFunction
from signify.authenticode import signed_pe

logger = logging.getLogger(__name__)

Range = collections.namedtuple("Range", "start end")
"""A range with a start and an end."""


class Finger:
    """A Finger defines how to hash a file to get specific fingerprints.

    The Finger contains one or more hash functions, a set of ranges in the file that
    are to be processed with these hash functions, and a description.

    While one Finger provides potentially multiple hashers, they all get fed the
    same ranges of the file.
    """

    def __init__(
        self, hashers: list[hashlib._Hash], ranges: list[Range], description: str
    ):
        """

        :param hashers: A list of hashers to feed.
        :param ranges: A list of Ranges that are hashed.
        :param description: The description of this Finger.
        """

        self._ranges = ranges
        self.hashers = hashers
        self.description = description

    @property
    def current_range(self) -> Range | None:
        """The working range of this Finger. Returns None if there is none."""
        if self._ranges:
            return self._ranges[0]
        return None

    def consume(self, start: int, end: int) -> None:
        """Consumes an entire range, or part thereof.

        If the finger has no ranges left, or the current range start is higher than
        the end of the consumed block, nothing happens. Otherwise, the current range is
        adjusted for the consumed block, or removed, if the entire block is consumed.
        For things to work, the consumed range and the current finger starts must be
        equal, and the length of the consumed range may not exceed the length of the
        current range.

        :param start: Beginning of range to be consumed.
        :param end: First offset after the consumed range (end + 1).
        :raises RuntimeError: if the start position of the consumed range is higher
            than the start of the current range in the finger, or if the consumed
            range cuts across block boundaries.
        """

        old = self.current_range
        if old is None:
            return

        if old.start > start:
            if old.start < end:
                raise RuntimeError("Block end too high.")
            return
        if old.start < start:
            raise RuntimeError("Block start too high.")
        if old.end == end:
            del self._ranges[0]
        elif old.end > end:
            self._ranges[0] = Range(end, old.end)
        else:
            raise RuntimeError("Block length exceeds range.")

    def update(self, block: bytes) -> None:
        """Given a data block, feed it to all the registered hashers."""

        for hasher in self.hashers:
            hasher.update(block)


class Fingerprinter:
    def __init__(self, file_obj: BinaryIO, block_size: int = 1000000):
        """A Fingerprinter is an interface to generate hashes of (parts) of a file.

        It is passed in a file object and given a set of :class:`Finger` s that define
        how a file must be hashed. It is a generic approach to not hashing parts of a
        file.

        :param file_obj: A file opened in bytes-mode
        :param block_size: The block size used to feed to the hashers.
        """

        self.file = file_obj
        self.block_size = block_size

        self.file.seek(0, os.SEEK_END)
        self._filelength = self.file.tell()

        self._fingers: list[Finger] = []

    def add_hashers(
        self,
        *hashers: HashFunction,
        ranges: list[Range] | None = None,
        description: str = "generic",
    ) -> None:
        """Add hash methods to the fingerprinter.

        :param hashers: A list of hashers to add to the Fingerprinter. This generally
            will be hashlib functions.
        :param ranges: A list of Range objects that the hashers should hash. If set
            to :const:`None`, it is set to the entire file.
        :param description: The name for the hashers. This name will return in
            :meth:`hashes`
        """
        concrete_hashers = [x() for x in hashers]
        if not ranges:
            ranges = [Range(0, self._filelength)]

        finger = Finger(concrete_hashers, ranges, description)
        self._fingers.append(finger)

    @property
    def _next_interval(self) -> Range | None:
        """Returns the next Range of the file that is to be hashed.

        For all fingers, inspect their next expected range, and return the
        lowest uninterrupted range of interest. If the range is larger than
        self.block_size, truncate it.

        :returns: Next range of interest in a Range namedtuple.
        """
        starts = {x.current_range.start for x in self._fingers if x.current_range}
        ends = {x.current_range.end for x in self._fingers if x.current_range}

        if not starts:
            return None

        min_start = min(starts)
        starts.remove(min_start)
        ends |= starts

        min_end = min(ends)
        if min_end - min_start > self.block_size:
            min_end = min_start + self.block_size
        return Range(min_start, min_end)

    def _hash_block(self, block: bytes, start: int, end: int) -> None:
        """_HashBlock feeds data blocks into the hashers of fingers.

        This function must be called before adjusting fingers for next
        interval, otherwise the lack of remaining ranges will cause the
        block not to be hashed for a specific finger.

        Start and end are used to validate the expected ranges, to catch
        unexpected use of that logic.

        :param block: The data block.
        :param start: Beginning offset of this block.
        :param offset: Offset of the next byte after the block.
        :raises RuntimeError: If the provided and expected ranges don't match.
        """
        for finger in self._fingers:
            expected_range = finger.current_range
            if expected_range is None:
                continue
            if (
                start > expected_range.start
                or (start == expected_range.start and end > expected_range.end)
                or (start < expected_range.start < end)
            ):
                raise RuntimeError("Cutting across fingers.")
            if start == expected_range.start:
                finger.update(block)

    def _consume(self, start: int, end: int) -> None:
        for finger in self._fingers:
            finger.consume(start, end)

    def hashes(self) -> dict[str, dict[str, bytes]]:
        """Finalizing function for the Fingerprint class.

        This method applies all the different hash functions over the previously
        specified different ranges of the input file, and computes the resulting hashes.

        After calling this function, the state of the object is reset to its  initial
        state, with no fingers defined.

        :returns: A dict of dicts, the outer dict being a mapping of the description
            (as set in :meth:`add_hashers` and the inner dict being a mapping of hasher
            name to digest.
        :raises RuntimeError: when internal inconsistencies occur.
        """
        while True:
            interval = self._next_interval
            if interval is None:
                break
            self.file.seek(interval.start, os.SEEK_SET)
            block = self.file.read(interval.end - interval.start)
            if len(block) != interval.end - interval.start:
                raise RuntimeError("Short read on file.")
            self._hash_block(block, interval.start, interval.end)
            self._consume(interval.start, interval.end)

        results = {}
        for finger in self._fingers:
            leftover = finger.current_range
            if leftover and (
                len(finger._ranges) > 1
                or leftover.start != self._filelength
                or leftover.end != self._filelength
            ):
                raise RuntimeError("Non-empty range remains.")

            res = {}
            for hasher in finger.hashers:
                res[hasher.name] = hasher.digest()
            results[finger.description] = res

        # Clean out things for a fresh start (on the same file object).
        self._fingers = []
        return results

    def hash(self) -> dict[str, bytes]:
        """Very similar to :meth:`hashes`, but only returns a single dict of hash names
        to digests.

        This method can only be called when the :meth:`add_hashers` method was called
        exactly once.
        """
        hashes = self.hashes()
        if len(hashes) != 1:
            raise RuntimeError("Can't return a single hash, use hashes() instead")

        return next(iter(hashes.values()))


class AuthenticodeFingerprinter(Fingerprinter):
    """An extension of the :class:`Fingerprinter` class that enables the calculation of
    authentihashes of PE Files.
    """

    def add_authenticode_hashers(self, *hashers: HashFunction) -> bool:
        """Specialized method of :meth:`add_hashers` to add hashers with ranges limited
        to those that are needed to calculate the hash of signed PE Files.
        """

        pefile = signed_pe.SignedPEFile(self.file)
        omit = pefile.get_authenticode_omit_sections()

        if omit is None:
            return False

        ranges = []
        start = 0
        for start_length in sorted(omit.values()):
            ranges.append(Range(start, start_length.start))
            start = sum(start_length)
        ranges.append(Range(start, self._filelength))

        self.add_hashers(*hashers, ranges=ranges, description="authentihash")
        return True


def main(*filenames: str) -> None:
    for filename in filenames:
        print(f"{filename}:")
        with pathlib.Path(filename).open("rb") as file_obj:
            fingerprinter = AuthenticodeFingerprinter(file_obj)
            fingerprinter.add_hashers(
                hashlib.md5, hashlib.sha1, hashlib.sha256, hashlib.sha512
            )
            fingerprinter.add_authenticode_hashers(
                hashlib.md5, hashlib.sha1, hashlib.sha256
            )
            results = fingerprinter.hashes()

            for description, result in sorted(results.items()):
                print(f"  {description or 'generic'}:")

                for k, v in sorted(result.items()):
                    if k == "_":
                        continue
                    print(f"    {k:<10}: {binascii.hexlify(v).decode('ascii')}")


if __name__ == "__main__":
    main(*sys.argv[1:])



# This is a derivative, modified, work from the verify-sigs project.
# Please refer to the LICENSE file in the distribution for more
# information. Original filename: generate_test_data.py
#
# Parts of this file are licensed as follows:
#
# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import io
import json
import os.path
import pathlib
import unittest

from signify.fingerprinter import (
    AuthenticodeFingerprinter,
    Finger,
    Fingerprinter,
    Range,
)

root_dir = pathlib.Path(__file__).parent


class FingerPrinterTestCase(unittest.TestCase):
    maxDiff = None

    def test_entire_blobs(self):
        for filename in (root_dir / "test_data").iterdir():
            if str(filename).endswith(".res") or not os.path.exists(
                str(filename) + ".res"
            ):
                continue
            with self.subTest(filename):
                with open(str(filename), "rb") as file_obj:
                    fingerprinter = AuthenticodeFingerprinter(file_obj)
                    fingerprinter.add_hashers(
                        hashlib.md5, hashlib.sha1, hashlib.sha256, hashlib.sha512
                    )
                    fingerprinter.add_authenticode_hashers(
                        hashlib.md5, hashlib.sha1, hashlib.sha256
                    )
                    results = fingerprinter.hashes()

                # convert to hex
                for v in results.values():
                    for k, b in v.items():
                        v[k] = b.hex()

                with open(str(filename) + ".res") as res_obj:
                    expected_results = json.load(res_obj)

                self.assertDictEqual(expected_results, results)

    def test_reasonable_interval(self):
        # Check if the limit on maximum blocksize for processing still holds.
        dummy = io.BytesIO(b"")
        fp = Fingerprinter(dummy)
        fp._fingers.append(Finger([], [Range(0, 1000001)], ""))

        start, stop = fp._next_interval
        self.assertEqual(0, start)
        self.assertEqual(1000000, stop)

    def test_adjustments(self):
        fp = Fingerprinter(io.BytesIO(b""))
        fp._fingers.append(Finger([], [Range(10, 20)], ""))

        # The remaining range should not yet be touched...
        fp._consume(9, 10)
        self.assertEqual([Range(10, 20)], fp._fingers[0]._ranges)
        # Trying to consume into the range. Blow up.
        self.assertRaises(RuntimeError, fp._consume, 9, 11)
        # We forgot a byte. Blow up.
        self.assertRaises(RuntimeError, fp._consume, 11, 12)
        # Consume a byte
        fp._consume(10, 11)
        self.assertEqual([Range(11, 20)], fp._fingers[0]._ranges)
        # Consumed too much. Blow up.
        self.assertRaises(RuntimeError, fp._consume, 11, 21)
        # Consume exactly.
        fp._consume(11, 20)
        self.assertEqual(0, len(fp._fingers[0]._ranges))

    def test_hash_block(self):
        # Does it invoke a hash function?
        dummy = b"12345"
        fp = Fingerprinter(io.BytesIO(dummy))
        big_finger = Finger([], [Range(0, len(dummy))], "")

        class MockHasher:
            def __init__(self):
                self.seen = b""

            def update(self, content):  # pylint: disable-msg=C6409
                self.seen += content

        hasher = MockHasher()

        big_finger.hashers = [hasher]
        fp._fingers.append(big_finger)
        # Let's process the block
        fp._hash_block(dummy, 0, len(dummy))
        self.assertEqual(hasher.seen, dummy)



from __future__ import annotations

import datetime
import hashlib
import pathlib
from typing import Any, Iterable

import mscerts
from asn1crypto import cms
from typing_extensions import Self

from signify import asn1
from signify._typing import HashFunction
from signify.asn1.hashing import _get_digest_algorithm
from signify.exceptions import (
    CertificateTrustListParseError,
    CTLCertificateVerificationError,
)
from signify.pkcs7 import signeddata
from signify.x509 import certificates, context

AUTHROOTSTL_PATH = pathlib.Path(mscerts.where(stl=True))


class CertificateTrustList(signeddata.SignedData):
    """A subclass of :class:`signify.pkcs7.SignedData`, containing a list of trusted
    root certificates. It is based on the following ASN.1 structure::

        CertificateTrustList ::= SEQUENCE {
          version CTLVersion DEFAULT v1,
          subjectUsage SubjectUsage,
          listIdentifier ListIdentifier OPTIONAL,
          sequenceNumber HUGEINTEGER OPTIONAL,
          ctlThisUpdate ChoiceOfTime,
          ctlNextUpdate ChoiceOfTime OPTIONAL,
          subjectAlgorithm AlgorithmIdentifier,
          trustedSubjects TrustedSubjects OPTIONAL,
          ctlExtensions [0] EXPLICIT Extensions OPTIONAL
        }
        CTLVersion ::= INTEGER {v1(0)}
        SubjectUsage ::= EnhancedKeyUsage
        ListIdentifier ::= OCTETSTRING
        TrustedSubjects ::= SEQUENCE OF TrustedSubject
        TrustedSubject ::= SEQUENCE{
          subjectIdentifier SubjectIdentifier,
          subjectAttributes Attributes OPTIONAL
        }
        SubjectIdentifier ::= OCTETSTRING

    .. attribute:: data

       The underlying ASN.1 data object

    .. attribute:: subject_usage

       Defines the EKU of the Certificate Trust List. Should be 1.3.6.1.4.1.311.20.1.

    .. attribute:: list_identifier

       The name of the template of the list.

    .. attribute:: sequence_number

       The unique number of this list

    .. attribute:: this_update

       The date of the current CTL.

    .. attribute:: next_update

       The date of the next CTL.

    .. attribute:: subject_algorithm

       Digest algorithm of verifying the list.

    .. warning::

       The CTL itself is currently not verifiable.

    """

    _expected_content_type = "microsoft_ctl"

    subject_usage: list[str]
    list_identifier: bytes | None
    sequence_number: int
    this_update: datetime.datetime | None
    next_update: datetime.datetime | None
    subject_algorithm: HashFunction

    def _parse(self) -> None:
        super()._parse()

        self.subject_usage = self.content["subject_usage"].native
        self.list_identifier = self.content["list_identifier"].native
        self.sequence_number = self.content["sequence_number"].native
        self.this_update = self.content["ctl_this_update"].native
        self.next_update = self.content["ctl_next_update"].native
        self.subject_algorithm = _get_digest_algorithm(
            self.content["subject_algorithm"],
            location="CertificateTrustList.subjectAlgorithm",
        )
        self._subjects = {}
        for subj in (
            CertificateTrustSubject(subject)
            for subject in self.content["trusted_subjects"]
        ):
            self._subjects[subj.identifier.hex().lower()] = subj
        # TODO: extensions??

    @property
    def subjects(self) -> Iterable[CertificateTrustSubject]:
        """A list of :class:`CertificateTrustSubject` s in this list."""

        return self._subjects.values()

    def verify_trust(
        self, chain: list[certificates.Certificate], *args: Any, **kwargs: Any
    ) -> bool:
        """Checks whether the specified certificate is valid in the given conditions
        according to this Certificate Trust List.

        :param List[Certificate] chain: The certificate chain to verify
        """

        # Find the subject belonging to this certificate
        subject = self.find_subject(chain[0])
        if not subject:
            raise CTLCertificateVerificationError(
                f"The root {chain[0]} is not in the certificate trust list"
            )
        return subject.verify_trust(chain, *args, **kwargs)

    def find_subject(
        self, certificate: certificates.Certificate
    ) -> CertificateTrustSubject | None:
        """Finds the :class:`CertificateTrustSubject` belonging to the provided
        :class:`signify.x509.Certificate`.

        :param signify.x509.Certificate certificate: The certificate to look for.
        :rtype: CertificateTrustSubject
        """

        if self.subject_algorithm == hashlib.sha1:
            identifier = certificate.sha1_fingerprint
        elif self.subject_algorithm == hashlib.sha256:
            identifier = certificate.sha256_fingerprint
        else:
            raise CertificateTrustListParseError(
                "The specified subject algorithm is not yet supported."
            )

        return self._subjects.get(identifier)

    @classmethod
    def from_stl_file(cls, path: pathlib.Path = AUTHROOTSTL_PATH) -> Self:
        """Loads a :class:`CertificateTrustList` from a specified path."""

        with path.open("rb") as f:
            data = cms.ContentInfo.load(f.read())

        if data["content_type"].native != "signed_data":
            raise CertificateTrustListParseError(
                "ContentInfo does not contain SignedData"
            )

        signed_data = cls(data["content"])
        return signed_data


class CertificateTrustSubject:
    """A subject listed in a :class:`CertificateTrustList`. The structure in this object
    has mostly been reverse-engineered using Windows tooling such as ``certutil -dump``.
    We do not pretend to have a complete picture of all the edge-cases that are
    considered.

    .. attribute:: data

       The underlying ASN.1 data object

    .. attribute:: attributes

       A dictionary mapping of attribute types to values.

    The following values are extracted from the attributes:

    .. attribute:: extended_key_usages

       Defines the EKU's the certificate is valid for. It may be empty, which we take
       as 'all is acceptable'.

    .. attribute:: friendly_name

       The friendly name of the certificate.

    .. attribute:: key_identifier

       The sha1 fingerprint of the certificate.

    .. attribute:: subject_name_md5

       The md5 of the subject name.

    .. attribute:: auth_root_sha256

       The sha256 fingerprint of the certificate.

    .. attribute:: disallowed_filetime

       The time since when a certificate has been disabled. Digital signatures with a
       timestamp prior to this date continue to be valid, but use cases after this date
       are prohibited. It may be used in conjunction with
       :attr:`disallowed_extended_key_usages` to define specific EKU's to be disabled.

    .. attribute:: root_program_chain_policies

       A list of EKU's probably used for EV certificates.

    .. attribute:: disallowed_extended_key_usages

       Defines the EKU's the certificate is not valid for. When used in combination with
       :attr:`disallowed_filetime`, the disabled EKU's are only disabled from that date
       onwards, otherwise, it means since the beginning of time.

    .. attribute:: not_before_filetime

       The time since when new certificates from this CA are not trusted. Certificates
       from prior the date will continue to validate. When used in conjunction with
       :attr:`not_before_extended_key_usages`, this only concerns certificates issued
       after this date for the defined EKU's.

    .. attribute:: not_before_extended_key_usages

       Defines the EKU's for which the :attr:`not_before_filetime` is considered. If
       that attribute is not defined, we assume that it means since the beginning of
       time.

    .. warning::

       The interpretation of the various attributes and their implications has been
       reverse-engineered. Though we seem to have a fairly solid understanding, various
       edge-cases may not have been considered.

    """

    extended_key_usages: list[str] | None
    friendly_name: str | None
    key_identifier: bytes
    subject_name_md5: bytes
    auth_root_sha256: bytes
    disallowed_filetime: datetime.datetime | None
    root_program_chain_policies: list[str] | None
    disallowed_extended_key_usages: list[str] | None
    not_before_filetime: datetime.datetime | None
    not_before_extended_key_usages: list[str] | None

    def __init__(self, data: asn1.ctl.TrustedSubject):
        self.data = data
        self._parse()

    def _parse(self) -> None:
        self.identifier = self.data["subject_identifier"].native
        self.attributes = {
            attr["type"].native: attr["values"].native[0]
            for attr in self.data["subject_attributes"]
        }

        self.extended_key_usages = self.attributes.get(
            "microsoft_ctl_enhkey_usage", None
        )
        self.friendly_name = self.attributes.get("microsoft_ctl_friendly_name", None)
        self.key_identifier = self.attributes.get("microsoft_ctl_key_identifier", b"")
        self.subject_name_md5 = self.attributes.get(
            "microsoft_ctl_subject_name_md5_hash", b""
        )
        # TODO: RootProgramCertPolicies not implemented
        self.auth_root_sha256 = self.attributes.get(
            "microsoft_ctl_auth_root_sha256_hash", b""
        )
        self.disallowed_filetime = self.attributes.get(
            "microsoft_ctl_disallowed_filetime", None
        )
        self.root_program_chain_policies = self.attributes.get(
            "microsoft_ctl_root_program_chain_policies", None
        )
        self.disallowed_extended_key_usages = self.attributes.get(
            "microsoft_ctl_disallowed_enhkey_usage", None
        )
        self.not_before_filetime = self.attributes.get(
            "microsoft_ctl_not_before_filetime", None
        )
        self.not_before_extended_key_usages = self.attributes.get(
            "microsoft_ctl_not_before_enhkey_usage", None
        )

    def verify_trust(
        self,
        chain: list[certificates.Certificate],
        context: context.VerificationContext,
    ) -> bool:
        """Checks whether the specified certificate is valid in the given conditions
        according to this Certificate Trust List.

        :param List[Certificate] chain: The certificate chain to verify.
        :param VerificationContext context: The context to verify with. Mainly the
            timestamp and extended_key_usages are used.
        """

        timestamp = context.timestamp
        if timestamp is None:
            timestamp = datetime.datetime.now(datetime.timezone.utc)
        extended_key_usages = context.extended_key_usages
        if extended_key_usages is None:
            extended_key_usages = ()

        # Start by converting the list of provided extended_key_usages to a list of OIDs
        requested_extended_key_usages = set(extended_key_usages)

        # Now check each of the properties
        if self.extended_key_usages and (
            requested_extended_key_usages - set(self.extended_key_usages)
        ):
            raise CTLCertificateVerificationError(
                f"The root {self.friendly_name} lists its extended key usages, but"
                f" {requested_extended_key_usages - set(self.extended_key_usages)} are"
                " not present"
            )

        # The notBefore time does concern the validity of the certificate that is being
        # validated. It must have a notBefore of before the timestamp
        if self.not_before_filetime is not None:
            to_verify_timestamp = chain[-1].valid_from

            if to_verify_timestamp >= self.not_before_filetime:
                # If there is a notBefore time, and there is no NotBeforeEnhkeyUsage,
                # then the validity concerns the entire certificate.
                if self.not_before_extended_key_usages is None:
                    raise CTLCertificateVerificationError(
                        f"The root {self.friendly_name} is disallowed for certificate"
                        f" issued after {self.not_before_filetime} (certificate is"
                        f" {to_verify_timestamp})"
                    )
                elif any(
                    eku in self.not_before_extended_key_usages
                    for eku in requested_extended_key_usages
                ):
                    raise CTLCertificateVerificationError(
                        f"The root {self.friendly_name} disallows requested EKU's"
                        f" {requested_extended_key_usages} to certificates issued after"
                        f" {self.not_before_filetime} (certificate is"
                        f" {to_verify_timestamp})"
                    )
        elif self.not_before_extended_key_usages is not None and any(
            eku in self.not_before_extended_key_usages
            for eku in requested_extended_key_usages
        ):
            raise CTLCertificateVerificationError(
                f"The root {self.friendly_name} disallows requested EKU's"
                f" {requested_extended_key_usages}"
            )

        # The DisallowedFiletime time does concern the timestamp of the signature
        # being verified.
        if self.disallowed_filetime is not None:
            if timestamp >= self.disallowed_filetime:
                # If there is a DisallowedFiletime, and there is no
                # DisallowedEnhkeyUsage, then the validity concerns the entire
                # certificate.
                if self.disallowed_extended_key_usages is None:
                    raise CTLCertificateVerificationError(
                        f"The root {self.friendly_name} is disallowed since "
                        f"{self.disallowed_filetime} (requested {timestamp})"
                    )
                elif any(
                    eku in self.disallowed_extended_key_usages
                    for eku in requested_extended_key_usages
                ):
                    raise CTLCertificateVerificationError(
                        f"The root {self.friendly_name} is disallowed for EKU's"
                        f" {self.disallowed_extended_key_usages} since"
                        f" {self.disallowed_filetime} (requested"
                        f" {requested_extended_key_usages} at {timestamp})"
                    )
        elif self.disallowed_extended_key_usages is not None and any(
            eku in self.disallowed_extended_key_usages
            for eku in requested_extended_key_usages
        ):
            raise CTLCertificateVerificationError(
                f"The root {self.friendly_name} disallows requested EKU's"
                f" {requested_extended_key_usages}"
            )

        return True



import hashlib
import unittest

from asn1crypto.algos import DigestAlgorithm
from asn1crypto.core import Boolean

from signify.exceptions import ParseError
from signify.pkcs7.signerinfo import _get_digest_algorithm


class GetDigestAlgorithmTest(unittest.TestCase):
    def test_acceptable_oid(self):
        sha1 = DigestAlgorithm()
        sha1["algorithm"] = "1.3.14.3.2.26"
        self.assertEqual(_get_digest_algorithm(sha1, location="test"), hashlib.sha1)

    def test_unknown_oid(self):
        invalid = DigestAlgorithm()
        invalid["algorithm"] = "1.2"
        self.assertRaises(ParseError, _get_digest_algorithm, invalid, location="test")

    def test_non_hashlib_oid(self):
        invalid = DigestAlgorithm()
        invalid["algorithm"] = "1.2.840.113549.1.9.3"
        self.assertRaises(ParseError, _get_digest_algorithm, invalid, location="test")

    def test_unacceptable_oid(self):
        sha1 = DigestAlgorithm()
        sha1["algorithm"] = "1.3.14.3.2.26"
        self.assertRaises(
            ParseError,
            _get_digest_algorithm,
            sha1,
            location="test",
            acceptable=["md5"],
        )

    def test_null_parameters(self):
        sha1 = DigestAlgorithm()
        sha1["algorithm"] = "1.3.14.3.2.26"
        sha1["parameters"] = "\x05\0"  # null value
        self.assertEqual(_get_digest_algorithm(sha1, location="test"), hashlib.sha1)

    def test_non_null_parameters(self):
        with self.assertRaises(TypeError):
            sha1 = DigestAlgorithm()
            sha1["algorithm"] = "1.3.14.3.2.26"
            sha1["parameters"] = Boolean(True)
            _get_digest_algorithm(sha1, location="test")



import glob
import os
from setuptools import setup


base_dir = os.path.dirname(__file__)
about = {}
with open(os.path.join(base_dir, "signify", "__init__.py")) as f:
    exec(f.read(), about)

try:
    long_description = open("README.rst", "r").read()
except Exception:
    long_description = None


setup(
    name="signify",
    version=about["__version__"],
    packages=[
        "signify",
        "signify.asn1",
        "signify.authenticode",
        "signify.pkcs7",
        "signify.x509",
    ],
    package_data={"signify": ["*.pem", "py.typed"]},
    include_package_data=True,
    url="https://github.com/ralphje/signify",
    download_url="https://github.com/ralphje/signify/tarball/v" + about["__version__"],
    license="MIT",
    author="Ralph Broenink",
    author_email="ralph@ralphbroenink.net",
    description="Module to generate and verify PE signatures",
    long_description=long_description,
    install_requires=[
        "certvalidator>=0.11",
        "asn1crypto>=1.3,<2",
        "oscrypto>=1.1,<2",
        "mscerts",
        "typing_extensions>=4.6.0",
    ],
    keywords=["authenticode", "authentihash", "fingerprinter", "pe"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Legal Industry",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Security :: Cryptography",
        "Topic :: System :: Software Distribution",
        "Topic :: Utilities",
    ],
)



from __future__ import annotations

import collections
import datetime
import logging
import re
from functools import cached_property
from typing import Any, ClassVar, Iterable, Iterator, cast, overload

import asn1crypto.pem
import asn1crypto.x509
from asn1crypto import cms
from oscrypto import asymmetric

from signify import x509
from signify._typing import HashFunction
from signify.exceptions import CertificateVerificationError

logger = logging.getLogger(__name__)


AlgorithmIdentifier = collections.namedtuple(
    "AlgorithmIdentifier", "algorithm parameters"
)


class Certificate:
    """Representation of a Certificate. It is built from an ASN.1 structure.

    .. attribute:: data

       The underlying ASN.1 data object

    .. attribute:: signature_algorithm
                   signature_value
                   subject_public_algorithm
                   subject_public_key

       These values are considered part of the certificate, but not
       fully parsed.

    .. attribute:: version

       This is the version of the certificate

    .. attribute:: serial_number

       The full integer serial number of the certificate

    .. attribute:: issuer
                   subject

       The :class:`CertificateName` for the issuer and subject.

    .. attribute:: valid_from
                   valid_to

       The datetime objects between which the certificate is valid.

    .. attribute:: extensions

       This is a list of extension objects.
    """

    signature_algorithm: Any
    signature_value: Any
    version: str
    serial_number: int
    issuer: CertificateName
    valid_from: datetime.datetime
    valid_to: datetime.datetime
    subject: CertificateName
    subject_public_algorithm: AlgorithmIdentifier
    subject_public_key: bytes
    extensions: dict[str, Any]

    def __init__(
        self,
        data: asn1crypto.x509.Certificate | cms.CertificateChoices,
    ):
        """

        :type data: asn1.pkcs7.ExtendedCertificateOrCertificate or
            asn1.x509.Certificate or asn1.x509.TBSCertificate
        :param data: The ASN.1 structure
        """

        self.data = data
        self._parse()

    def _parse(self) -> None:
        if isinstance(self.data, cms.ExtendedCertificate):
            raise NotImplementedError(
                "Support for extendedCertificate is not implemented"
            )

        if isinstance(self.data, cms.CertificateChoices):
            if self.data.name != "certificate":
                raise NotImplementedError(
                    f"This is not a certificate, but a {self.data.name}"
                )
            self.data = self.data.chosen

        self.signature_algorithm = self.data["signature_algorithm"].native
        self.signature_value = self.data["signature_value"].native
        tbs_certificate = self.data["tbs_certificate"]

        self.version = tbs_certificate["version"].native
        self.serial_number = tbs_certificate["serial_number"].native
        self.issuer = CertificateName(tbs_certificate["issuer"])
        self.valid_from = tbs_certificate["validity"]["not_before"].native
        self.valid_to = tbs_certificate["validity"]["not_after"].native
        self.subject = CertificateName(tbs_certificate["subject"])

        self.subject_public_algorithm = AlgorithmIdentifier(
            algorithm=tbs_certificate["subject_public_key_info"]["algorithm"][
                "algorithm"
            ].native,
            parameters=tbs_certificate["subject_public_key_info"]["algorithm"][
                "parameters"
            ].native,
        )
        self.subject_public_key = tbs_certificate["subject_public_key_info"][
            "public_key"
        ].dump()

        self.extensions = {}
        if tbs_certificate["extensions"].native is not None:
            for extension in tbs_certificate["extensions"]:
                self.extensions[extension["extn_id"].native] = extension[
                    "extn_value"
                ].native

    def __str__(self) -> str:
        return (
            f"{self.subject.dn}"
            f" (serial:{self.serial_number}, sha1:{self.sha1_fingerprint})"
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.issuer,
                self.serial_number,
                self.subject,
                self.subject_public_algorithm,
                self.subject_public_key,
            )
        )

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Certificate)
            and self.issuer == other.issuer
            and self.serial_number == other.serial_number
            and self.subject == other.subject
            and self.subject_public_algorithm == other.subject_public_algorithm
            and self.subject_public_key == other.subject_public_key
        )

    @classmethod
    def from_der(cls, content: bytes) -> Certificate:
        """Load the Certificate object from DER-encoded data"""
        return cls(asn1crypto.x509.Certificate.load(content))

    @classmethod
    def from_pem(cls, content: bytes) -> Certificate:
        """Reads a Certificate from a PEM formatted file."""
        return next(cls.from_pems(content))

    @classmethod
    def from_pems(cls, content: bytes) -> Iterator[Certificate]:
        """Reads a Certificate from a PEM formatted file."""
        for _type_name, _headers, der_bytes in asn1crypto.pem.unarmor(
            content, multiple=True
        ):
            yield cls.from_der(der_bytes)

    @cached_property
    def to_der(self) -> bytes:
        """Returns the DER-encoded data from this certificate."""
        return cast(bytes, self.data.dump())

    @cached_property
    def sha256_fingerprint(self) -> str:
        return cast(str, self.data.sha256_fingerprint).replace(" ", "").lower()

    @cached_property
    def sha1_fingerprint(self) -> str:
        return cast(str, self.data.sha1_fingerprint).replace(" ", "").lower()

    def verify_signature(
        self,
        signature: bytes,
        data: bytes,
        algorithm: HashFunction,
        allow_legacy: bool = False,
    ) -> None:
        """Verifies whether the signature bytes match the data using the hashing
        algorithm. Supports RSA and EC keys. Note that not all hashing algorithms
        are supported.

        :param bytes signature: The signature to verify
        :param bytes data: The data that must be verified
        :type algorithm: a hashlib function
        :param algorithm: The hashing algorithm to use
        :param bool allow_legacy: If True, allows a legacy signature verification.
            This method is intended for the case where the encryptedDigest does not
            contain an ASN.1 structure, but a raw hash value instead. It is attempted
            automatically when verification of the RSA signature fails.

            This case is described in more detail on
            https://mta.openssl.org/pipermail/openssl-users/2015-September/002053.html
        """

        public_key = asymmetric.load_public_key(self.data.public_key)
        if public_key.algorithm == "rsa":
            verify_func = asymmetric.rsa_pkcs1v15_verify
        elif public_key.algorithm == "dsa":
            verify_func = asymmetric.dsa_verify
        elif public_key.algorithm == "ec":
            verify_func = asymmetric.ecdsa_verify
        else:
            raise CertificateVerificationError(
                f"Signature algorithm {public_key.algorithm} is unsupported for {self}"
            )

        try:
            verify_func(public_key, signature, data, algorithm().name)
        except Exception as e:
            if not allow_legacy or public_key.algorithm != "rsa":
                raise CertificateVerificationError(f"Invalid signature for {self}: {e}")
        else:
            return

        try:
            hasher = algorithm()
            hasher.update(data)
            asymmetric.rsa_pkcs1v15_verify(
                public_key, signature, hasher.digest(), "raw"
            )
        except Exception as e:
            raise CertificateVerificationError(
                f"Invalid signature for {self} (legacy attempted): {e}"
            )

    def potential_chains(
        self, context: x509.VerificationContext
    ) -> Iterator[list[Certificate]]:
        """Alias for :meth:`VerificationContext.potential_chains`"""

        return context.potential_chains(self)

    def verify(self, context: x509.VerificationContext) -> Iterable[Certificate]:
        """Alias for :meth:`VerificationContext.verify`"""

        return context.verify(self)


class CertificateName:
    OID_TO_RDN: ClassVar[dict[str, str]] = {
        "2.5.4.3": "CN",  # common name
        "2.5.4.6": "C",  # country
        "2.5.4.7": "L",  # locality
        "2.5.4.8": "ST",  # stateOrProvince
        "2.5.4.9": "STREET",  # street
        "2.5.4.10": "O",  # organization
        "2.5.4.11": "OU",  # organizationalUnit
        "0.9.2342.19200300.100.1.25": "DC",  # domainComponent
        "1.2.840.113549.1.9.1": "EMAIL",  # emailaddress
    }

    def __init__(self, data: asn1crypto.x509.Name | asn1crypto.x509.GeneralName):
        if isinstance(data, asn1crypto.x509.GeneralName):
            if data.name != "directory_name":
                raise NotImplementedError(
                    f"CertificateNames of type {data.name} not supported"
                )
            data = data.chosen
        self.data = data

    def __eq__(self, other: object) -> bool:
        return isinstance(other, CertificateName) and self.rdns == other.rdns

    def __hash__(self) -> int:
        return hash(self.rdns)

    def __str__(self) -> str:
        return self.dn

    @property
    def dn(self) -> str:
        """Returns an (almost) rfc2253 compatible string given a RDNSequence"""

        result = []
        for type, value in self.get_components():
            #   If the AttributeType is in a published table of attribute types
            #   associated with LDAP [4], then the type name string from that table
            #   is used, otherwise it is encoded as the dotted-decimal encoding of
            #   the AttributeType's OBJECT IDENTIFIER.

            # Escaping according to RFC2253
            value = re.sub('([,+"<>;\\\\])', r"\\\1", value)
            if value.startswith("#"):
                value = "\\" + value
            if value.endswith(" "):
                value = value[:-1] + "\\ "
            result.append(f"{type}={value}")
        return ", ".join(result)

    @property
    def rdns(self) -> Iterable[tuple[str, str]]:
        """A list of all components of the object."""
        return tuple(self.get_components())

    @overload
    def get_components(
        self, component_type: None = None
    ) -> Iterator[tuple[str, str]]: ...

    @overload
    def get_components(self, component_type: str) -> Iterator[str]: ...

    def get_components(
        self, component_type: str | None = None
    ) -> Iterator[tuple[str, str]] | Iterator[str]:
        """Get individual components of this CertificateName

        :param component_type: if provided, yields only values of this type,
            if not provided, yields tuples of (type, value)
        """

        for n in list(self.data.chosen)[::-1]:
            type_value = n[0]  # get the AttributeTypeAndValue object

            type = self.OID_TO_RDN.get(
                type_value["type"].dotted, type_value["type"].dotted
            )
            value = type_value["value"].native

            if component_type is not None:
                if component_type in (
                    type_value["type"].dotted,
                    type_value["type"].native,
                    type,
                ):
                    yield value
            else:
                yield type, value



import pathlib
import unittest

from signify.authenticode.authroot import CertificateTrustList

root_dir = pathlib.Path(__file__).parent


class AuthrootTestCase(unittest.TestCase):
    def test_certificate_can_be_opened(self):
        ctl = CertificateTrustList.from_stl_file()
        # assume at least 400 items in the list
        self.assertGreaterEqual(len(ctl.subjects), 400)



from __future__ import annotations

from typing import Any, Sequence, cast

from asn1crypto import cms
from asn1crypto.core import Asn1Value
from typing_extensions import Self

from signify._typing import HashFunction
from signify.asn1.hashing import _get_digest_algorithm
from signify.exceptions import ParseError
from signify.pkcs7 import signerinfo
from signify.x509.certificates import Certificate
from signify.x509.context import CertificateStore


class SignedData:
    """A generic SignedData object. The SignedData object is defined in RFC2315 and
    RFC5652 (amongst others) and defines data that is signed by one or more signers.

    It is based on the following ASN.1 object (as per RFC2315)::

        SignedData ::= SEQUENCE {
          version Version,
          digestAlgorithms DigestAlgorithmIdentifiers,
          contentInfo ContentInfo,
          certificates [0] IMPLICIT ExtendedCertificatesAndCertificates OPTIONAL,
          crls [1] IMPLICIT CertificateRevocationLists OPTIONAL,
          signerInfos SignerInfos
        }

    This class supports RFC2315 and RFC5652.

    .. attribute:: data

       The underlying ASN.1 data object

    .. attribute:: digest_algorithm

       The digest algorithm, i.e. the hash algorithm, that is used by the signers of
       the data.

    .. attribute:: content_type

       The class of the type of the content in the object.

    .. attribute:: content

       The actual content, as parsed by the :attr:`content_type` spec.

    .. attribute:: certificates
       :type: CertificateStore

       A list of all included certificates in the SignedData. These can be used to
       determine a valid validation path from the signer to a root certificate.

    .. attribute:: signer_infos
       :type: List[SignerInfo]

       A list of all included SignerInfo objects
    """

    data: cms.SignedData
    digest_algorithm: HashFunction
    content_type: str
    content: Asn1Value
    certificates: CertificateStore
    signer_infos: Sequence[signerinfo.SignerInfo]

    _expected_content_type: str | None = None
    _signerinfo_class: type[signerinfo.SignerInfo] | str | None = None

    def __init__(self, data: cms.SignedData):
        """

        :param data: The ASN.1 structure of the SignedData object
        """

        if isinstance(self._signerinfo_class, str):
            self._signerinfo_class = globals()[self._signerinfo_class]

        self.data = data
        self._parse()

    @classmethod
    def from_envelope(cls, data: bytes, *args: Any, **kwargs: Any) -> Self:
        """Loads a :class:`SignedData` object from raw data that contains ContentInfo.

        :param data: The bytes to parse
        """
        content_info = cms.ContentInfo.load(data)

        if content_info["content_type"].native != "signed_data":
            raise ParseError("ContentInfo does not contain SignedData")

        signed_data = cls(content_info["content"], *args, **kwargs)
        return signed_data

    def _parse(self) -> None:
        #         ('version', CMSVersion),
        #         ('digest_algorithms', DigestAlgorithms),
        #         ('encap_content_info', None),
        #         ('certificates', CertificateSet, {'implicit': 0, 'optional': True}),
        #         ('crls', RevocationInfoChoices, {'implicit': 1, 'optional': True}),
        #         ('signer_infos', SignerInfos),

        # digestAlgorithms
        if len(self.data["digest_algorithms"]) != 1:
            raise ParseError(
                f"SignedData.digestAlgorithms must contain"
                f" exactly 1 algorithm, not {len(self.data['digestAlgorithms'])}"
            )
        self.digest_algorithm = _get_digest_algorithm(
            self.data["digest_algorithms"][0], "SignedData.digestAlgorithm"
        )

        self.content_type = self.data["encap_content_info"]["content_type"].native

        if self.content_type != self._expected_content_type:
            raise ParseError(
                f"SignedData.contentInfo contains {self.content_type},"
                f" expected {self._expected_content_type}"
            )

        self._real_content = self.data["encap_content_info"]["content"]
        if hasattr(self._real_content, "parsed"):
            self.content = self._real_content.parsed
        else:
            self.content = self._real_content

        # Certificates
        self.certificates = CertificateStore(
            [
                Certificate(cert)
                for cert in self.data["certificates"]
                if not isinstance(cert, cms.CertificateChoices)
                or cert.name == "certificate"
            ]
        )

        # SignerInfo
        if self._signerinfo_class is not None:
            assert not isinstance(self._signerinfo_class, str)
            self.signer_infos = [
                self._signerinfo_class(si, parent=self)
                for si in self.data["signer_infos"]
            ]

    @property
    def content_digest(self) -> bytes:
        """Returns the digest of the content of the SignedData object,
        adhering to the specs in RFC2315, 9.3; the identifier (tag) and
        length need to be stripped for hashing.
        """

        if hasattr(self._real_content, "parsed"):
            # Handle the case where the content is a ParsableOctetString, and
            # self.content.contents may refer to its children
            hash_content = bytes(self._real_content)
        else:
            hash_content = self.content.contents

        blob_hasher = self.digest_algorithm()
        blob_hasher.update(hash_content)
        return blob_hasher.digest()
